import logging
from pprint import pprint
from time import time
from pathlib import Path
from typing import Callable
import warnings
import pytest
from numbers import Real, Integral, Number

import numpy as np
from pytest import raises

from sklearn.tree._criterion import MSE, MAE, FriedmanMSE, Poisson
from sklearn.tree._splitter import BestSplitter, RandomSplitter
from sklearn.utils.validation import check_random_state
from sklearn.utils._testing import assert_allclose
from sklearn.utils._param_validation import Interval, validate_params

from bipartite_learn.tree._splitter_factory import (
    make_bipartite_splitter,
    make_bipartite_ss_splitter,
    make_semisupervised_criterion,
)
from bipartite_learn.tree._bipartite_criterion import (
    GMO,
    GMOSA,
)
from bipartite_learn.tree._axis_criterion import (
    AxisSquaredError,
    AxisSquaredErrorGSO,
    AxisFriedmanGSO,
    AxisGini,
    AxisEntropy,
)
from bipartite_learn.tree._semisupervised_criterion import (
    SSCompositeCriterion,
)
from bipartite_learn.tree._unsupervised_criterion import (
    UnsupervisedSquaredError,
    UnsupervisedFriedman,
)
from .utils.make_examples import make_interaction_blobs, make_interaction_regression

from .utils.tree_utils import apply_monopartite_splitter, apply_bipartite_splitter
from .utils.test_utils import (
    parse_args, stopwatch, gen_mock_data, melt_2d_data, assert_equal_dicts,
)
from . import test_criteria


# Default test params
DEF_PARAMS = dict(
    seed=0,
    # shape=(10, 10),
    shape=(50, 50),
    # nattrs=(10, 9),
    nattrs=(10, 10),
    nrules=1,
    min_samples_leaf=1,
    # min_samples_leaf=100,
    transpose_test=False,
    noise=.5,
    inspect=False,
    plot=False,
    start=0,
    end=-1,
    verbose=False,
)


@pytest.fixture
def verbose():
    return True


@pytest.fixture(params=[(0, 0, 0, 0)])#, (5, 5, -5, -5)])
def dataset_slice(request):
    # start_row, start_col, end_row, end_col
    return request.param


@pytest.fixture(params=range(10))
def random_state(request):
    return check_random_state(request.param)


@pytest.fixture
def dataset(random_state):
    XX, Y, x, y = make_interaction_regression(
        n_samples=(50, 40),
        n_features=(10, 9),
        n_targets=None,
        min_target=0.0,
        max_target=100.0,
        noise=10.0,
        return_molten=True,
        return_tree=False,
        random_state=random_state,
        max_depth=None,
    )
    return XX, Y, x, y


@pytest.fixture
def dataset_mono_semisupervised(dataset, supervision):
    XX, Y, x, y = dataset
    return XX, Y, x, np.hstack((x, y))


@pytest.fixture(params=[0.0, 0.001, 0.1, 0.2328, 0.569, 0.782, 0.995, 1.0])
def supervision(request):
    return request.param


@pytest.fixture(
    params=[
        (0.0, 0.0), (0.1, 0.9), (0.2328, 0.569), (0.782, 0.782), (1.0, 1.0),
    ],
)
def bipartite_supervision(request):
    return request.param


def compare_splitters_1d2d_ideal(
    splitter1,
    splitter2,
    tol=0,
    **params,
):
    params = DEF_PARAMS | dict(noise=0) | params

    if params['noise']:
        warnings.warn(f"noise={params['noise']}. Setting it to zero"
                      " since noise=0 is what defines an ideal split.")
        params['noise'] = 0

    result1, result2 = compare_splitters_1d2d(
        splitter1, splitter2, tol, **params)

    assert result1['improvement'] != 0
    assert result1['impurity_left'] == 0
    assert result1['impurity_right'] == 0

    assert result2['improvement'] != 0
    assert result2['impurity_left'] == 0
    assert result2['impurity_right'] == 0


def monopartite_splitter_factory(
    x,
    y,
    *,
    random_state,
    splitter,
    criterion,
):
    """Returns a factory to make a splitter based on the dataset provided.
    """
    return splitter(
        criterion=criterion(
            n_outputs=y.shape[1],
            n_samples=x.shape[0],
        ),
        max_features=x.shape[1],
        min_samples_leaf=1,
        min_weight_leaf=0.0,
        random_state=random_state,
    )


def monopartite_ss_splitter_factory(
    x,
    y,
    *,
    random_state,
    splitter,
    supervised_criterion,
    unsupervised_criterion,
    supervision,
    ss_adapter=None,
):

    criterion = make_semisupervised_criterion(
        supervision=supervision,
        supervised_criterion=supervised_criterion,
        unsupervised_criterion=unsupervised_criterion,
        n_features=x.shape[1],
        n_samples=x.shape[0],
        n_outputs=y.shape[1],
        ss_class=ss_adapter,
    )
    criterion.set_X(x.astype('float64'))

    return splitter(
        criterion=criterion,
        max_features=x.shape[1],
        min_samples_leaf=1,
        min_weight_leaf=0.0,
        random_state=random_state,
    )


def bipartite_splitter_factory(
    x, y,
    *,
    random_state,
    **kwargs,
):
    return make_bipartite_splitter(
        max_features=[X.shape[1] for X in x],
        n_samples=y.shape,
        n_outputs=y.shape[::-1],
        random_state=random_state,
        **kwargs,
    )


def bipartite_ss_splitter_factory(
    x, y,
    *,
    random_state,
    **kwargs,
):
    splitter = make_bipartite_ss_splitter(
        max_features=[X.shape[1] for X in x],
        n_samples=y.shape,
        n_outputs=y.shape[::-1],
        n_features=[X.shape[1] for X in x],
        random_state=random_state,
        **kwargs,
    )
    splitter.criterion_wrapper.set_X(
        x[0].astype('float64'),
        x[1].astype('float64'),
    )
    return splitter


@pytest.mark.parametrize(
    (
        'splitter1_factory, splitter1_args,'
        'splitter2_factory, splitter2_args'
    ), [
        (
            monopartite_splitter_factory, {
                'splitter': BestSplitter,
                'criterion': MSE,
            },
            monopartite_splitter_factory, {
                'splitter': BestSplitter,
                'criterion': MSE,
            },
        ),
    ],
    ids=['mse_ss_sup'],
)
def test_compare_1d_splitters(
    splitter1_factory,
    splitter1_args,
    splitter2_factory,
    splitter2_args,
    dataset,
    dataset_slice,
    random_state,
):
    XX, Y, *_ = dataset
    X = XX[0]
    start, _, end, _ = dataset_slice

    splitter1 = splitter1_factory(
        X, Y,
        random_state=random_state,
        **splitter1_args,
    )
    splitter2 = splitter2_factory(
        X, Y,
        random_state=random_state,
        **splitter2_args,
    )

    with stopwatch(
        f'Testing first splitter ({splitter1.__class__.__name__})...'
    ):
        result1 = apply_monopartite_splitter(
            splitter1, X, Y,
            start=start,
            end=end,
            verbose=verbose,
        )
    print('Best split 1 found:')
    pprint(result1)

    with stopwatch(
        f'Testing second splitter ({splitter2.__class__.__name__})...'
    ):
        result2 = apply_monopartite_splitter(
            splitter2, X, Y,
            start=start,
            end=end,
            verbose=verbose,
        )
    print('Best split 2 found:')
    pprint(result2)

    assert_equal_dicts(result1, result2)


@pytest.mark.parametrize(
    (
        'mono_splitter_factory, mono_splitter_args,'
        'bi_splitter_factory, bi_splitter_args'
    ), [
        (
            monopartite_splitter_factory, {
                'splitter': BestSplitter,
                'criterion': MSE,
            },
            bipartite_splitter_factory, {
                'splitters': BestSplitter,
                'criteria': AxisSquaredErrorGSO,
                'bipartite_criterion_class': GMOSA,
            },
        ),
        # Friedman impurity depends on weighted_n_left/right, predictable
        # failing here.
    ],
    ids=['mse'],
)
def test_compare_1d2d_splitters_gso(
    mono_splitter_factory,
    mono_splitter_args,
    bi_splitter_factory,
    bi_splitter_args,
    dataset,
    dataset_slice,
    random_state,
):
    XX, Y, x, y = dataset
    start_row, start_col, end_row, end_col = dataset_slice
    # Columns subset not supported.
    start_col, end_col = 0, 0

    mono_splitter = mono_splitter_factory(
        x, y,
        random_state=random_state,
        **mono_splitter_args,
    )
    bi_splitter = bi_splitter_factory(
        XX, Y,
        random_state=random_state,
        **bi_splitter_args,
    )

    with stopwatch(
        f'Testing monopartite splitter ({mono_splitter.__class__.__name__})...'
    ):
        result_monopartite = apply_monopartite_splitter(
            mono_splitter,
            x, y,
            start=start_row * Y.shape[1],
            end=end_row * Y.shape[1],
            verbose=verbose,
        )
    print('Best monopartite split found:')
    pprint(result_monopartite)

    # Adapt values to compare
    n_rows, n_cols = Y.shape
    result_monopartite['axis'] = (
        int(result_monopartite['feature'] >= XX[0].shape[1])
    )
    pos = result_monopartite['pos']

    # FIXME: It's complicated whem we use start_row != 0
    result_monopartite['pos'] = (
        pos // n_rows if result_monopartite['axis'] == 1 else pos // n_cols
    )

    print('Corrected monopartite split:')
    pprint(result_monopartite)

    with stopwatch(
        f'Testing bipartite splitter ({bi_splitter.__class__.__name__})...'
    ):
        result_bipartite = apply_bipartite_splitter(
            bi_splitter,
            XX, Y,
            start=[start_row, start_col],
            end=[end_row, end_col],
            verbose=verbose,
        )
    print('Best bipartite split found:')
    pprint(result_bipartite)

    # NOTE: Sometimes a diferent feature yields the same partitioning
    # (especially with few samples in the dataset).
    assert_equal_dicts(
        result_monopartite,
        result_bipartite,
        ignore=['pos'], # Complicated if start_row != 0. Threshold is enough.
    )


@pytest.mark.parametrize(
    (
        'mono_splitter_factory, mono_splitter_args,'
        'bi_splitter_factory, bi_splitter_args'
    ), [
        (
            monopartite_splitter_factory, {
                'splitter': BestSplitter,
                'criterion': MSE,
            },
            bipartite_splitter_factory, {
                'splitters': BestSplitter,
                'criteria': AxisSquaredError,
                'bipartite_criterion_class': GMOSA,
            },
        ),
        (
            monopartite_splitter_factory, {
                'splitter': BestSplitter,
                'criterion': FriedmanMSE,
            },
            bipartite_splitter_factory, {
                'splitters': BestSplitter,
                'criteria': AxisFriedmanGSO,
                'bipartite_criterion_class': GMOSA,
            },
        ),
    ],
    ids=['mse', 'friedman'],
)
def test_compare_1d2d_splitters_gmo(
    mono_splitter_factory,
    mono_splitter_args,
    bi_splitter_factory,
    bi_splitter_args,
    dataset,
    dataset_slice,
    random_state,
):
    XX, Y, x, y = dataset
    start_row, start_col, end_row, end_col = dataset_slice

    mono_splitter_rows = mono_splitter_factory(
        XX[0], Y,
        random_state=random_state,
        **mono_splitter_args,
    )
    mono_splitter_cols = mono_splitter_factory(
        XX[1], Y.T,
        random_state=random_state,
        **mono_splitter_args,
    )

    bi_splitter = bi_splitter_factory(
        XX, Y,
        random_state=random_state,
        **bi_splitter_args,
    )

    with stopwatch(
        f'Testing 2D splitter ({bi_splitter.__class__.__name__})...'
    ):
        result_bipartite = apply_bipartite_splitter(
            bi_splitter,
            XX, Y,
            start=[start_row, start_col],
            end=[end_row, end_col],
            verbose=verbose,
        )
    print('Best bipartite split found:')
    pprint(result_bipartite)

    with stopwatch(
        f'Testing 1D splitter ({type(mono_splitter_rows).__name__}) on rows...'
    ):
        result_rows = apply_monopartite_splitter(
            mono_splitter_rows,
            XX[0], Y,
            start=start_row,
            end=end_row,
            verbose=verbose,
        )
    result_rows['axis'] = 0
    print('Best monopartite rows split found:')
    pprint(result_rows)

    with stopwatch(
        f'Testing 1D splitter ({type(mono_splitter_cols).__name__}) on columns...'
    ):
        result_columns = apply_monopartite_splitter(
            mono_splitter_cols,
            XX[1], np.ascontiguousarray(Y.T),
            start=start_col,
            end=end_col,
            verbose=verbose,
        )
    result_columns['axis'] = 1
    result_columns['feature'] += XX[0].shape[1]
    print('Best monopartite columns split found:')
    pprint(result_columns)

    results_monopartite = (result_rows, result_columns)
    best_axis = result_bipartite['axis']

    assert (
        results_monopartite[best_axis]['improvement'] >
        results_monopartite[1 - best_axis]['improvement']
    ), 'Wrong axis.'

    assert_equal_dicts(
        results_monopartite[best_axis],
        result_bipartite,
        subset={'pos', 'feaure', 'threshold'},
    )


@pytest.mark.parametrize(
    (
        'splitter1_factory, splitter1_args,'
        'splitter2_factory, splitter2_args'
    ), [
        (
            monopartite_splitter_factory, {
                'splitter': BestSplitter,
                'criterion': MSE,
            },
            monopartite_splitter_factory, {
                'splitter': BestSplitter,
                'criterion': MSE,
            },
        ),
    ],
    ids=['mse'],
)
def test_compare_1d_splitters_ss(
    splitter1_factory,
    splitter1_args,
    splitter2_factory,
    splitter2_args,
    dataset,
    dataset_slice,
    supervision,
    random_state,
):
    if splitter1_args.get('supervision', False) is None:
        splitter1_args['supervision'] = supervision

    if splitter2_args.get('supervision', False) is None:
        splitter2_args['supervision'] = supervision
    
    X, Y, x, y = dataset
    X[0] *= (1 - supervision)
    Y *= supervision
    dataset = X, Y, x, y

    test_compare_1d_splitters(
        splitter1_factory,
        splitter1_args,
        splitter2_factory,
        splitter2_args,
        dataset,
        dataset_slice,
        random_state=random_state,
    )


@pytest.mark.parametrize(
    (
        'mono_splitter_factory, mono_splitter_args,'
        'bi_splitter_factory, bi_splitter_args'
    ), [
        (
            monopartite_ss_splitter_factory, {
                'splitter': BestSplitter,
                'supervised_criterion': UnsupervisedSquaredError,
                'unsupervised_criterion': UnsupervisedSquaredError,
                'supervision': None,
            },
            bipartite_ss_splitter_factory, {
                'splitters': BestSplitter,
                'supervised_criteria': AxisSquaredErrorGSO,
                'unsupervised_criteria': UnsupervisedSquaredError,
                'supervision': None,
            },
        ),
        (
            monopartite_ss_splitter_factory, {
                'splitter': BestSplitter,
                'supervised_criterion': UnsupervisedFriedman,
                'unsupervised_criterion': UnsupervisedSquaredError,
                'supervision': None,
            },
            bipartite_ss_splitter_factory, {
                'splitters': BestSplitter,
                'supervised_criteria': AxisFriedmanGSO,
                'unsupervised_criteria': UnsupervisedSquaredError,
                'supervision': None,
            },
        )
    ],
    ids=['mse', 'friedman'],
)
def test_compare_1d2d_splitters_gso_ss(
    mono_splitter_factory,
    mono_splitter_args,
    bi_splitter_factory,
    bi_splitter_args,
    dataset,
    dataset_slice,
    bipartite_supervision,
    random_state,
):
    if mono_splitter_args.get('supervision', False) is None:
        mono_splitter_args['supervision'] = bipartite_supervision[0]

    if bi_splitter_args.get('supervision', False) is None:
        bi_splitter_args['supervision'] = bipartite_supervision

    test_compare_1d2d_splitters_gso(
        mono_splitter_factory,
        mono_splitter_args,
        bi_splitter_factory,
        bi_splitter_args,
        dataset,
        dataset_slice,
        random_state=random_state,
    )


@pytest.mark.parametrize(
    (
        'mono_splitter_factory, mono_splitter_args,'
        'bi_splitter_factory, bi_splitter_args'
    ), [
        (
            monopartite_ss_splitter_factory, {
                'splitter': BestSplitter,
                'supervised_criterion': UnsupervisedFriedman,
                'unsupervised_criterion': UnsupervisedSquaredError,
                'supervision': None,
            },
            bipartite_ss_splitter_factory, {
                'splitters': BestSplitter,
                'supervised_criteria': AxisFriedmanGSO,
                'unsupervised_criteria': UnsupervisedSquaredError,
                'supervision': None,
            },
        ),
        (
            monopartite_ss_splitter_factory, {
                'splitter': BestSplitter,
                'supervised_criterion': UnsupervisedSquaredError,
                'unsupervised_criterion': UnsupervisedSquaredError,
                'supervision': None,
            },
            bipartite_ss_splitter_factory, {
                'splitters': BestSplitter,
                'supervised_criteria': AxisSquaredError,
                'unsupervised_criteria': UnsupervisedSquaredError,
                'supervision': None,
            },
        ),
    ],
    ids=['friedman', 'mse'],
)
def test_compare_1d2d_splitters_gmo_ss(
    mono_splitter_factory,
    mono_splitter_args,
    bi_splitter_factory,
    bi_splitter_args,
    dataset,
    dataset_slice,
    bipartite_supervision,
    random_state,
):
    if mono_splitter_args.get('supervision', False) is None:
        mono_splitter_args['supervision'] = bipartite_supervision[0]

    if bi_splitter_args.get('supervision', False) is None:
        bi_splitter_args['supervision'] = bipartite_supervision

    test_compare_1d2d_splitters_gmo(
        mono_splitter_factory,
        mono_splitter_args,
        bi_splitter_factory,
        bi_splitter_args,
        dataset,
        dataset_slice,
        random_state=random_state,
    )


@pytest.mark.skip
def test_1d2d_ideal(**params):
    params = DEF_PARAMS | params
    return compare_splitters_1d2d_ideal(
        splitter1=BestSplitter,
        splitter2=BestSplitter,
        **params,
    )


@pytest.mark.skip
def test_ss_1d2d_unsup(**params):
    params = DEF_PARAMS | params
    ss2d_splitter = make_bipartite_ss_splitter(
        splitters=BestSplitter,
        ss_criteria=SSCompositeCriterion,
        supervised_criteria=MSE,
        unsupervised_criteria=MSE,
        supervision=0.,
        max_features=params['nattrs'],
        n_features=params['nattrs'],
        n_samples=params['shape'],
        n_outputs=1,
    )

    return compare_splitters_1d2d(
        splitter1=BestSplitter,
        splitter2=ss2d_splitter,
        unsupervised_1d=True,
        **params,
    )


@pytest.mark.skip
def test_ss_1d2d(supervision, **params):
    """Compare 1D to 2D version of semisupervised MSE splitter.
    """
    params = DEF_PARAMS | params

    splitter1 = BestSplitter(
        criterion=make_semisupervised_criterion(
            supervision=supervision,
            supervised_criterion=MSE,
            unsupervised_criterion=MSE,
            n_features=np.sum(params['nattrs']),
            n_samples=np.prod(params['shape']),
            n_outputs=1,
        ),
        max_features=np.sum(params['nattrs']),
        min_samples_leaf=params['min_samples_leaf'],
        min_weight_leaf=0.,
        random_state=check_random_state(params['seed']),
    )

    ss2d_splitter = make_bipartite_ss_splitter(
        splitters=BestSplitter,
        ss_criteria=SSCompositeCriterion,
        supervised_criteria=MSE,
        unsupervised_criteria=MSE,
        supervision=supervision,
        max_features=params['nattrs'],
        n_features=params['nattrs'],
        n_samples=params['shape'],
        n_outputs=1,
        random_state=params['seed'],
        min_samples_leaf=params['min_samples_leaf'],
        min_weight_leaf=0.,
    )

    return compare_splitters_1d2d(
        splitter1=splitter1,
        splitter2=ss2d_splitter,
        semisupervised_1d=True,
        **params,
    )


@pytest.mark.skip
def test_ss_1d2d_ideal_split(**params):
    params = DEF_PARAMS | params
    ss2d_splitter = make_bipartite_ss_splitter(
        splitters=BestSplitter,
        ss_criteria=SSCompositeCriterion,
        supervised_criteria=MSE,
        unsupervised_criteria=MSE,
        supervision=1.,
        max_features=params['nattrs'],
        n_features=params['nattrs'],
        n_samples=params['shape'],
        n_outputs=1,
    )

    return compare_splitters_1d2d_ideal(
        splitter1=BestSplitter,
        splitter2=ss2d_splitter,
        **params,
    )


@pytest.mark.skip
def test_ss_1d2d_blobs(supervision, random_state, **params):
    """Test axis decision-semisupervision

    Assert that axis decision semisupervision chooses the split based only on
    supervised data, integrating unsupervised score after it is found.
    """
    params = DEF_PARAMS | params
    n_row_features = params['nattrs'][0]

    X, Y, x, y = make_interaction_blobs(
        return_molten=True,
        n_features=params['nattrs'],
        n_samples=params['shape'],
        random_state=check_random_state(random_state),
        noise=5.0,
        centers=10,
    )

    splitter1 = BestSplitter(
        criterion=make_semisupervised_criterion(
            supervised_criterion=MSE,
            unsupervised_criterion=MSE,
            n_samples=y.shape[0],
            n_outputs=y.shape[1],
            n_features=x.shape[1],
            supervision=supervision,
        ),
        max_features=x.shape[1],
        min_samples_leaf=1,
        min_weight_leaf=0.0,
        random_state=check_random_state(random_state),
    )
    splitter2 = make_bipartite_ss_splitter(
        splitters=BestSplitter,
        supervised_criteria=MSE,
        unsupervised_criteria=MSE,
        supervision=supervision,
        max_features=params['nattrs'],
        n_features=params['nattrs'],
        n_samples=params['shape'],
        n_outputs=1,
        min_samples_leaf=1,
        min_weight_leaf=0.0,
        random_state=check_random_state(random_state),
        axis_decision_only=False,
    )
    result1 = apply_monopartite_splitter(
        splitter1,
        x,
        np.hstack((x, y)),
        verbose=params['verbose'],
    )
    ref_criterion = test_criteria.ReferenceSquaredError()
    ref_criterion.set_data(x, y, supervision=supervision)

    manual_result = ref_criterion.evaluate_split(
        pos=result1['pos'],
        feature=result1['feature'],
    )
    # rtol=1e-4 because x is converted from float32
    assert_equal_dicts(
        result1, manual_result, msg_prefix='(manual) ', rtol=1e-4,
    )

    n_rows, n_cols = Y.shape
    axis = result1['feature'] >= n_row_features
    result1['axis'] = int(axis)
    pos = result1['pos']
    result1['old_pos'] = pos
    result1['pos'] = pos // n_rows if axis else pos // n_cols

    result2 = apply_bipartite_splitter(
        splitter2,
        X,
        Y,
        verbose=params['verbose'],
    )
    assert_equal_dicts(result2, result1, rtol=1e-4)


@pytest.mark.skip
def apply_monopartite_splitter_gmo(**params):
    params = DEF_PARAMS | params
    BipartiteSplitter = make_bipartite_splitter(
        max_features=params['nattrs'],
        n_samples=params['shape'],
        n_outputs=params['shape'][::-1],
        min_samples_leaf=params['min_samples_leaf'],
        min_weight_leaf=0.0,
    )

    compare_splitters_1d2d_ideal(
        splitter1=BestSplitter,
        splitter2=BipartiteSplitter,
        multioutput_1d=True,
        **params,
    )


# TODO: compare results
@pytest.mark.skip
def apply_monopartite_splitter_gmo_classification(**params):
    params = DEF_PARAMS | params
    BipartiteSplitter = make_bipartite_splitter(
        bipartite_criterion_class=GMO,
        is_classification=True,
        splitters=BestSplitter,
        criteria=[AxisGini, AxisGini],
        max_features=params['nattrs'],
        n_samples=params['shape'],
        n_outputs=params['shape'][::-1],
        n_classes=[np.repeat(2, i) for i in params['shape'][::-1]],
        min_samples_leaf=params['min_samples_leaf'],
        min_weight_leaf=0.0,
    )

    with pytest.raises(AssertionError):
        compare_splitters_1d2d(
            splitter1=BestSplitter,
            splitter2=BipartiteSplitter,
            multioutput_1d=True,
            **params,
        )


@pytest.mark.skip
def test_ss_axis_decision_only(supervision, random_state, **params):
    """Test axis decision-semisupervision

    Assert that axis decision semisupervision chooses the split based only on
    supervised data, integrating unsupervised score after it is found.
    """
    if supervision == 0.0:
        pytest.skip(
            'axis decision only should be compared to supervised splitter'
        )

    params = DEF_PARAMS | params
    n_row_features = params['nattrs'][0]

    X, Y, x, y, gen_tree = make_interaction_regression(
        return_molten=True,
        return_tree=True,
        n_features=params['nattrs'],
        n_samples=params['shape'],
        random_state=random_state,
        max_depth=1,
        max_target=100,
        noise=0.1,
    )
    y = y.reshape((-1, 1))

    splitter2 = make_bipartite_ss_splitter(
        splitters=BestSplitter,
        supervised_criteria=MSE,
        unsupervised_criteria=MSE,
        supervision=supervision,
        max_features=params['nattrs'],
        n_features=params['nattrs'],
        n_samples=params['shape'],
        n_outputs=1,
        random_state=random_state,
        min_samples_leaf=params['min_samples_leaf'],
        min_weight_leaf=0.0,
        axis_decision_only=True,
    )
    splitter2.criterion_wrapper.init_X(X[0], X[1])
    result2 = apply_bipartite_splitter(
        splitter2,
        X,
        Y,
        verbose=params['verbose'],
    )

    if result2['axis'] == 0:
        x = x[:, :n_row_features]
    else:  # axis == 1
        x = x[:, n_row_features:]

    splitter1 = BestSplitter(
        criterion=MSE(n_outputs=y.shape[1], n_samples=x.shape[0]),
        max_features=x.shape[1],
        min_samples_leaf=1,
        min_weight_leaf=0.0,
        random_state=check_random_state(random_state),
    )
    result1 = apply_monopartite_splitter(
        splitter1,
        x,
        y,
        verbose=params['verbose'],
    )
    ref_criterion = test_criteria.ReferenceSquaredError()
    ref_criterion.set_data(x, y, supervision=supervision)

    manual_result = ref_criterion.evaluate_split(
        x, y, pos=result1['pos'], feature=result1['feature'],
    )
    assert_equal_dicts(
        result1,
        manual_result,
        ignore={'threshold', 'feature'},
        msg_prefix='(reference) ',
        rtol=1e-4,
        atol=1e-8,
    )
    n_rows, n_cols = params['shape']
    result1['axis'] = result2['axis']
    pos = result1['pos']
    result1['pos'] = pos // n_rows if result1['axis'] else pos // n_cols

    if result1['axis'] == 1:
        result1['feature'] += n_row_features

    # Compare only feature and threshold
    assert_equal_dicts(
        result1,
        result2,
        ignore={
            'impurity_parent',
            'impurity_left',
            'impurity_right',
            'improvement',
        },
        rtol=1e-4,
        atol=1e-8,
    )


def test_gini_mse_identity(random_state, **params):
    params = DEF_PARAMS | params
    XX, Y = make_interaction_regression(
        n_samples=(40, 50),
        n_features=(9, 10),
        random_state=random_state,
    )

    # XX = [check_symmetric(X, raise_warning=False) for X in XX]
    Y = (Y > Y.mean()).astype('float64')  # Turn into binary.
    node_impurity = (Y.var(0).mean() + Y.var(1).mean()) / 2
    logging.info(f'parent_impurity: {node_impurity}')

    splitter_gini = make_bipartite_splitter(
        bipartite_criterion_class=GMO,
        splitters=BestSplitter,
        criteria=AxisGini,
        max_features=params['nattrs'],
        n_classes=[np.repeat(2, n) for n in params['shape'][::-1]],
        n_outputs=params['shape'][::-1],
        min_samples_leaf=params['min_samples_leaf'],
        min_weight_leaf=0.0,
    )
    splitter_mse = make_bipartite_splitter(
        bipartite_criterion_class=GMO,
        splitters=BestSplitter,
        criteria=AxisSquaredError,
        max_features=params['nattrs'],
        n_samples=params['shape'],
        n_outputs=params['shape'][::-1],
        min_samples_leaf=params['min_samples_leaf'],
        min_weight_leaf=0.0,
    )
    result_mse = apply_bipartite_splitter(
        splitter_mse, XX, Y,
        verbose=params['verbose'],
    )
    result_gini = apply_bipartite_splitter(
        splitter_gini, XX, Y,
        verbose=params['verbose'],
    )

    result_gini['improvement'] /= 2
    result_gini['impurity_left'] /= 2
    result_gini['impurity_right'] /= 2
    result_gini['impurity_parent'] /= 2

    assert_equal_dicts(result_gini, result_mse)



"""Set of tools to apply standard estimators to bipartite datasets.

TODO: Docs.
TODO: check fit inputs.
"""
from __future__ import annotations
from typing import Callable, Sequence
import numpy as np
from sklearn.base import (
    BaseEstimator,
    TransformerMixin,
    MetaEstimatorMixin,
    clone,
    is_classifier,
    is_regressor,
)
from sklearn.utils.metaestimators import available_if
from sklearn.utils._tags import _safe_tags
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError
from sklearn.utils import check_random_state
from imblearn.base import BaseSampler
from .melter import melt_multipartite_dataset
from .utils import check_multipartite_params
from .base import (
    BaseMultipartiteEstimator,
    BaseBipartiteEstimator,
    BaseMultipartiteSampler,
)
from .utils import _X_is_multipartite

__all__ = [
    "GlobalSingleOutputWrapper",
    "LocalMultiOutputWrapper",
    "MultipartiteTransformerWrapper",
    "MultipartiteSamplerWrapper",
]


def _estimator_has(attr):
    """Check that primary estimators has `attr`.

    Used together with `avaliable_if` in `GlobalSingleOutputWrapper`.
    """

    def check(self):
        # raise original `AttributeError` if `attr` does not exist
        getattr(self.estimator, attr)
        return True

    return check


def _secondary_estimators_have(attr):
    """Check that primary estimators has `attr`.

    Used together with `avaliable_if` in `LocalMultiOutputWrapper`.
    """

    def check(self):
        # raise original `AttributeError` if `attr` does not exist
        getattr(self.secondary_rows_estimator, attr)
        getattr(self.secondary_cols_estimator, attr)
        return True

    return check


def _primary_estimators_have(attr):
    """Check that secondary estimators have `attr`.

    Used together with `avaliable_if` in `LocalMultiOutputWrapper`.
    """

    def check(self):
        # raise original `AttributeError` if `attr` does not exist
        getattr(self.primary_rows_estimator, attr)
        getattr(self.primary_cols_estimator, attr)
        return True

    return check


class IncompatibleEstimatorsError(ValueError):
    """Raised when user tries to wrap incompatible estimators."""


class GlobalSingleOutputWrapper(BaseMultipartiteEstimator, MetaEstimatorMixin):
    """Employ the GSO strategy to adapt sstandard estimators to bipartite data.

    In this strategy, the estimator is applied to concatenations of a feature
    vector from the first sample domain with a feature vector from the second
    domain, while `y` is considered a unidimensional vector.

    .. image:: _static/user_guide/gso.svg
        :align: center
        :width: 50%
        :class: only-light

    .. image:: _static/user_guide/gso_dark.svg
        :align: center
        :width: 50%
        :class: only-dark

    Read more in the :ref:`User Guide <user_guide:global_single_output>`.

    References
    ----------
    .. [1] :doi:`Global Multi-Output Decision Trees for interaction prediction \
       <doi.org/10.1007/s10994-018-5700-x>`
       Pliakos, Geurts and Vens, 2018
    """

    def __init__(
        self,
        estimator: BaseEstimator,
        under_sampler: BaseSampler | None = None,
    ):
        """Initialize the wrapper.

        Parameters
        ----------
        estimator : BaseEstimator
            Estimator or transformer to receive the adapted bipartite data.

        under_sampler : BaseSampler or None, default=None
            Optional sampler to be applied to the transformed data before
            fitting the estimator. Useful in cases where using all the possible
            interactions is not computationally feasible.
        """
        self.estimator = estimator
        self.under_sampler = under_sampler

    def _melt_input(self, X, y=None):
        if not _X_is_multipartite(X):
            return X, y
        return melt_multipartite_dataset(X, y)

    def _fit_resample_input(self, X, y):
        if self.under_sampler is None:
            return X, y
        self.under_sampler_ = clone(self.under_sampler)
        return self.under_sampler_.fit_resample(X, y)

    def fit(self, X, y=None, **fit_params):
        X, y = self._validate_data(X, y, reset=False)
        Xt, yt = self._melt_input(X, y=y)
        Xt, yt = self._fit_resample_input(Xt, yt)
        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(X=Xt, y=yt, **fit_params)
        return self

    @available_if(_estimator_has("predict"))
    def predict(self, X, **predict_params):
        X, _ = self._melt_input(X)
        return self.estimator_.predict(X, **predict_params)

    @available_if(_estimator_has("fit_predict"))
    def fit_predict(self, X, y=None, **fit_params):
        X, y = self._validate_data(X, y, reset=False)
        Xt, yt = self._melt_input(X, y=y)
        Xt, yt = self._fit_resample_input(Xt, yt)
        self.estimator_ = clone(self.estimator)
        return self.estimator_.fit_predict(Xt, yt, **fit_params)

    @available_if(_estimator_has("predict_proba"))
    def predict_proba(self, X, **predict_proba_params):
        X, _ = self._melt_input(X)
        return self.estimator_.predict_proba(X, **predict_proba_params)

    @available_if(_estimator_has("decision_function"))
    def decision_function(self, X):
        X, _ = self._melt_input(X)
        return self.estimator_.decision_function(X)

    @available_if(_estimator_has("score_samples"))
    def score_samples(self, X):
        X, _ = self._melt_input(X)
        return self.estimator_.score_samples(X)

    @available_if(_estimator_has("predict_log_proba"))
    def predict_log_proba(self, X, **predict_log_proba_params):
        X, _ = self._melt_input(X)
        return self.estimator_.predict_log_proba(X, **predict_log_proba_params)

    @available_if(_estimator_has("transform"))
    def transform(self, X):
        Xt, _ = self._melt_input(X)
        return self.estimator_.transform(Xt)

    @available_if(_estimator_has("fit_transform"))
    def fit_transform(self, X, y=None, **fit_params):
        X, y = self._validate_data(X, y, reset=False)
        Xt, yt = self._melt_input(X, y=y)
        Xt, yt = self._fit_resample_input(Xt, yt)
        self.estimator_ = clone(self.estimator)
        return self.estimator_.fit_transform(Xt, yt, **fit_params)

    @available_if(_estimator_has("inverse_transform"))
    def inverse_transform(self, Xt):
        X, _ = self._melt_input(Xt)
        return self.estimator_.inverse_transform(X)

    @available_if(_estimator_has("score"))
    def score(self, X, y=None, sample_weight=None):
        X, y = self._melt_input(X, y=y)
        score_params = {}
        if sample_weight is not None:
            score_params["sample_weight"] = sample_weight
        return self.estimator_.score(X, y, **score_params)

    @property
    def _estimator_type(self):
        return getattr(self.estimator, "_estimator_type", None)

    @property
    def classes_(self):
        """The classes labels. Only exist if the estimator is a classifier."""
        return self.estimator_.classes_

    @property
    def n_features_in_(self):
        """Number of features seen during `fit`."""
        return self.estimator_.n_features_in_

    @property
    def feature_names_in_(self):
        """Names of features seen during `fit`."""
        return self.estimator_.feature_names_in_

    def __sklearn_is_fitted__(self):
        """Indicate whether the estimator has been fit."""
        return hasattr(self, "estimator_")


class LocalMultiOutputWrapper(BaseBipartiteEstimator):
    """Implements the Local Multi-Output strategy for adapting estimators.
    
    This wrapper facilitates the implementation of the local multi-output
    approach to adapt monopartite estimators to bipartite scenarios. In this
    approach, four multi-output estimators are aggregated.

    The training procedure (calling ``fit(X_train, y)``) consists simply of:

    1. Train a primary rows estimator on X_train[0] and y_train.
    2. Train a primary columns estimator on X_train[1] and y_train.T.

    The prediction procedure then utilities the predictions of the primary
    estimators in order to be able to make predictions on completely new
    interactions. ``predict(X_test)`` will perform the following steps:

    1. a. Use ``self.primary_cols_estimator_`` to predict new columns for
            the interaction matrix, that correspond to the targets of
            ``X_test[0]``.
    1. b. Use ``self.primary_rows_estimator_`` to predict new rows for the
            interaction matrix, that correspond to the targets of X_test[1].

    2. a. Fit the secondary rows estimator on the newly predicted columns
            and X_test[0].
    2. a. Fit the secondary columns estimator on the newly predicted rows
            and X_test[1].
    
    3. Combine the predictions of the secondary estimators using
        ``self.combine_predictions_func(rows_pred, cols_pred)``.

    If ``self.independent_labels`` is ``False``, then the original
    training data is appended to the training data of the secondary
    estimators in step 2, allowing the secondary estimators to explore
    inter-output correlations. 

    See the :ref:`User Guide <user_guide:local_multi_output>` for a diagram and
    more information.

    Attributes
    ----------
    primary_rows_estimator_ : BaseEstimator
        The fitted primary rows estimator.
    primary_cols_estimator_ : BaseEstimator
        The fitted primary columns estimator.
    secondary_rows_estimator_ : BaseEstimator
        The fitted secondary rows estimator.
    secondary_cols_estimator_ : BaseEstimator
        The fitted secondary columns estimator.

    Notes
    -----
    Note that the secondary estimators must be refit every time the
    wrapper's ``predict()`` method is called, which may increase prediction
    time depending on the type of secondary estimators chosen by the user.
    
    Compositions of single-output estimators can also be used
    instead of multi-output estimators, which can be implemented with
    scikit-learn wrappers such as :class:`MultiOutputRegressor` or
    :class:`MultiOutputClassifier`. This could be an interesting option in
    cases where the base estimators do not natively support multiple
    outputs.
        
    See Also
    --------
    GlobalSingleOutputWrapper : A wrapper that fits a single-output estimator
        to bipartite datasets.
    MultiOutputRegressor : A scikit-learn wrapper that fits a separate regressor
        for each output variable.
    MultiOutputClassifier : A scikit-learn wrapper that fits a separate
        classifier for each output variable.
    
    Examples
    --------

    .. code-block:: python

        from bipartite_learn.datasets import NuclearReceptorsLoader
        from bipartite_learn.wrappers import LocalMultiOutputWrapper
        from sklearn.svm import SVC
        from sklearn.neighbors import KNeighborsClassifier
        from sklearn.multioutput import MultiOutputClassifier
        
        X, y = NuclearReceptorsLoader().load()  # X is a list of two matrices
        bipartite_clf = LocalMultiOutputWrapper(
            primary_rows_estimator=MultiOutputClassifier(SVC()),
            primary_cols_estimator=MultiOutputClassifier(SVC()),
            secondary_rows_estimator=KNeighborsClassifier(),
            secondary_cols_estimator=KNeighborsClassifier(),
        )
        bipartite_clf.fit(X, y)
    
    References
    ----------
    .. [1] :doi:`Global Multi-Output Decision Trees for interaction prediction \
       <doi.org/10.1007/s10994-018-5700-x>`
       Pliakos, Geurts and Vens, 2018
    
    .. [2] :doi:`Yamanishi, Y., Araki, M., Gutteridge, A., Honda, W., &
           Kanehisa, M. (2008). Prediction of drug-target interaction networks
           from the integration of chemical and genomic spaces. Bioinformatics,
           24(13), i232–i240. <https://doi.org/10.1093%2Fbioinformatics%2Fbtn162>`
    """

    def __init__(
        self,
        primary_rows_estimator: BaseEstimator,
        primary_cols_estimator: BaseEstimator,
        secondary_rows_estimator: BaseEstimator,
        secondary_cols_estimator: BaseEstimator,
        combine_predictions_func: Callable = np.mean,
        combine_func_kwargs: dict | None = None,
        independent_labels: bool = True,
    ):
        """Initializes the wrapper.

        This wrapper facilitates the implementation of the local multi-output
        approach to adapt monopartite estimators to bipartite scenarios. In this
        approach, four multi-output estimators are aggregated.

        Parameters
        ----------
        primary_rows_estimator : BaseEstimator
            The estimator to be used as the primary rows estimator.
        primary_cols_estimator : BaseEstimator
            The estimator to be used as the primary columns estimator.
        secondary_rows_estimator : BaseEstimator
            The estimator to be used as the secondary rows estimator.
        secondary_cols_estimator : BaseEstimator
            The estimator to be used as the secondary columns estimator.
        combine_predictions_func : Callable, default=np.mean
            The function used to combine the predictions of the secondary
            estimators.
        combine_func_kwargs : dict, default=None
            A dictionary of keyword arguments to be passed to the
            ``combine_predictions_func`` function. If None, no keyword arguments
            are passed.
        independent_labels : bool, default=True
            If ``True``, the secondary estimators are trained on only the data
            predicted by the primary estimators. If ``False``, the original
            training data is appended to the training data of the secondary
            estimators, allowing the secondary estimators to explore
            inter-output correlations.

        See the :ref:`User Guide <local_multi_output>` for more information on
        the Local Multi-Output strategy for adapting estimators.
        """
        self.primary_rows_estimator = primary_rows_estimator
        self.primary_cols_estimator = primary_cols_estimator
        self.secondary_rows_estimator = secondary_rows_estimator
        self.secondary_cols_estimator = secondary_cols_estimator

        self.independent_labels = independent_labels
        self.combine_predictions_func = combine_predictions_func
        self.combine_func_kwargs = combine_func_kwargs

    def _check_estimators(self):
        for estimator in (
            self.primary_rows_estimator,
            self.primary_cols_estimator,
            self.secondary_rows_estimator,
            self.secondary_cols_estimator,
        ):
            # TODO: 'multioutput_only' tag should imply 'multioutput' tag
            if not (
                _safe_tags(estimator, "multioutput")
                or _safe_tags(estimator, "multioutput_only")
            ):
                raise IncompatibleEstimatorsError(
                    f"All estimators wrapped by {self.__class__.__name__} "
                    f"must support multioutput functionality but {estimator} "
                    "does not. Some meta-estimators defined in scikit-learn "
                    "may be useful, check https://scikit-learn.org/stable/"
                    "modules/multiclass.html"
                )

        if getattr(self.secondary_rows_estimator, "_estimator_type") != getattr(
            self.secondary_cols_estimator, "_estimator_type"
        ):
            raise IncompatibleEstimatorsError(
                "Secondary estimators must be of the same type (regressor"
                ", classifier, etc.). See https://scikit-learn.org/stable/"
                "developers/develop.html#estimator-types"
            )

        if _safe_tags(self.primary_rows_estimator, "pairwise") != _safe_tags(
            self.primary_cols_estimator, "pairwise"
        ):
            raise IncompatibleEstimatorsError(
                "Both or none of the primary estimators must be pairwise."
            )

        self.primary_rows_estimator_ = clone(self.primary_rows_estimator)
        self.primary_cols_estimator_ = clone(self.primary_cols_estimator)
        self.secondary_rows_estimator_ = clone(self.secondary_rows_estimator)
        self.secondary_cols_estimator_ = clone(self.secondary_cols_estimator)

    def fit(self, X, y, **fit_params):
        """Fits the wrapper to the training data.

        Raises
        ------
        IncompatibleEstimatorsError
            If any of the estimators passed as arguments does not support
            multi-output functionality.  If the secondary estimators are not of
            the same type (e.g., regressor, classifier).  If only one of the
            primary estimators is pairwise.
        """
        if not _X_is_multipartite(X):
            # TODO: specific error and function to check multipartite data
            # TODO: set estimator tag 'multipartite_only' to True
            raise ValueError("X must be multipartite.")
        self._check_estimators()
        X, y = self._validate_data(X, y, reset=False)
        self.X_fit_ = X

        if not self.independent_labels:
            self.y_fit_ = y

        self.primary_rows_estimator_.fit(X[0], y, **fit_params)
        self.primary_cols_estimator_.fit(X[1], y.T, **fit_params)

        return self

    def _secondary_fit(self, X):
        """Fit secondary estimators at prediction time."""
        check_is_fitted(self)

        # Transposed because they will be used as training labels.
        new_y_rows = self.primary_rows_estimator_.predict(X[0]).T
        new_y_cols = self.primary_cols_estimator_.predict(X[1]).T

        # If the secondary estimators are able to take advantage of correlated
        # labels, we train them with all the labels we have, even if most pre-
        # dicted columns will be discarded.
        if not self.independent_labels:
            new_y_cols = np.hstack((new_y_cols, self.y_fit_))
            new_y_rows = np.hstack((new_y_rows, self.y_fit_.T))

        self.secondary_rows_estimator_.fit(self.X_fit_[0], new_y_cols)
        self.secondary_cols_estimator_.fit(self.X_fit_[1], new_y_rows)

        return self

    def _combine_predictions(self, rows_pred, cols_pred):
        return np.apply_along_axis(
            func1d=self.combine_predictions_func,
            axis=0,
            arr=(rows_pred, cols_pred),
            **(self.combine_func_kwargs or {}),
        )

    @available_if(_secondary_estimators_have("predict"))
    def predict(self, X, **predict_params):
        if is_classifier(self):
            return self._predict_classification(X, **predict_params)
        elif is_regressor(self):
            return self._predict_regression(X, **predict_params)
        raise ValueError

    def _predict_regression(self, X, **predict_params):
        self._secondary_fit(X)

        rows_pred = self.secondary_rows_estimator_.predict(
            X[0],
            **predict_params,
        )
        cols_pred = self.secondary_cols_estimator_.predict(
            X[1],
            **predict_params,
        ).T

        if not self.independent_labels:
            # Get only the predictions corresponding to the instances being
            # predicted.
            rows_pred = rows_pred[:, : X[1].shape[0]]
            cols_pred = cols_pred[: X[0].shape[0]]

        return self._combine_predictions(rows_pred, cols_pred).reshape(-1)

    def _predict_classification(self, X, **predict_params):
        check_is_fitted(self)

        if hasattr(self, "predict_proba"):
            proba = self.predict_proba(X, **predict_params)
        elif hasattr(self, "predict_log_proba"):
            proba = self.predict_log_proba(X, **predict_params)
        elif hasattr(self, "decision_function"):
            proba = self.decision_function(X, **predict_params)
        else:
            raise ValueError(
                "If a classifier, the estimator must have a predict_proba,"
                "predict_log_proba or decision_function method."
            )

        # FIXME: we must ensure that the classes are the same for all rows
        # and columns of y, and that they are in the same order.
        return self.classes_[np.argmax(proba, axis=1)]

    # TODO: make also available if only a row or column estimator has
    @available_if(
        lambda self: (
            is_classifier(self)
            and (
                _primary_estimators_have("fit_predict")
                or _secondary_estimators_have("fit_predict")
            )
        )
    )
    def fit_predict(self, X, y=None, **fit_params):
        self._check_estimators()
        X, y = self._validate_data(X, y)
        self.X_fit_ = X

        if not self.independent_labels:
            self.y_fit_ = y

        # Transposed because they will be used as training labels.
        if hasattr(self.primary_rows_estimator_, "fit_predict"):
            new_y_rows = self.primary_rows_estimator_.fit_predict(
                X[0],
                y,
                **fit_params,
            ).T
            new_y_cols = self.primary_cols_estimator_.fit_predict(
                X[1],
                y.T,
                **fit_params,
            ).T
        else:
            self.fit(X, y, **fit_params)
            new_y_rows = self.primary_rows_estimator_.predict(X[0]).T
            new_y_cols = self.primary_cols_estimator_.predict(X[1]).T

        # If the secondary estimators are able to take advantage of correlated
        # labels, we train them with all the labels we have, even if most pre-
        # dicted columns will be discarded.
        if not self.independent_labels:
            new_y_cols = np.hstack((new_y_cols, y))
            new_y_rows = np.hstack((new_y_rows, y.T))

        if hasattr(self.secondary_rows_estimator_, "fit_predict"):
            rows_pred = self.secondary_rows_estimator_.fit_predict(
                X[0], new_y_cols, **fit_params
            )
            cols_pred = self.secondary_cols_estimator_.fit_predict(
                X[1], new_y_rows, **fit_params
            ).T
        else:
            self.secondary_rows_estimator_.fit(X[0], new_y_cols)
            self.secondary_cols_estimator_.fit(X[1], new_y_rows)
            rows_pred = self.secondary_rows_estimator_.predict(X[0])
            cols_pred = self.secondary_cols_estimator_.predict(X[1]).T

        if not self.independent_labels:
            # Get only the predictions corresponding to the instances being
            # predicted.
            rows_pred = rows_pred[:, : X[1].shape[0]]
            cols_pred = cols_pred[: X[0].shape[0]]

        return self._combine_predictions(rows_pred, cols_pred).reshape(-1)

    @available_if(_secondary_estimators_have("predict_proba"))
    def predict_proba(self, X, **predict_proba_params):
        self._secondary_fit(X)

        rows_pred = np.array(
            self.secondary_rows_estimator_.predict_proba(X[0], **predict_proba_params)
        )
        cols_pred = np.array(
            self.secondary_cols_estimator_.predict_proba(X[1], **predict_proba_params)
        ).transpose((1, 0, 2))

        if not self.independent_labels:
            # Get only the predictions corresponding to the instances being
            # predicted.
            rows_pred = rows_pred[:, : X[1].shape[0]]
            cols_pred = cols_pred[: X[0].shape[0]]

        # FIXME: Does not work in some cases, such as 'max()'
        return self._combine_predictions(rows_pred, cols_pred).reshape(-1, 2)

    @available_if(_secondary_estimators_have("predict_log_proba"))
    def predict_log_proba(self, X, **predict_log_proba_params):
        self._secondary_fit(X)

        rows_pred = np.array(
            self.secondary_rows_estimator_.predict_log_proba(
                X[0], **predict_log_proba_params
            )
        )
        cols_pred = np.array(
            self.secondary_cols_estimator_.predict_log_proba(
                X[1], **predict_log_proba_params
            )
        ).transpose((1, 0, 2))

        if not self.independent_labels:
            # Get only the predictions corresponding to the instances being
            # predicted.
            rows_pred = rows_pred[:, : X[1].shape[0]]
            cols_pred = cols_pred[: X[0].shape[0]]

        # FIXME: Does not work in some cases, such as 'max()'
        return self._combine_predictions(rows_pred, cols_pred).reshape(-1, 2)

    @available_if(_secondary_estimators_have("decision_function"))
    def decision_function(self, X, **decision_function_params):
        self._secondary_fit(X)

        rows_pred = self.secondary_rows_estimator_.decision_function(
            X[0], **decision_function_params
        )
        cols_pred = self.secondary_cols_estimator_.decision_function(
            X[1], **decision_function_params
        ).T

        if not self.independent_labels:
            # Get only the predictions corresponding to the instances being
            # predicted.
            rows_pred = rows_pred[:, : X[1].shape[0]]
            cols_pred = cols_pred[: X[0].shape[0]]

        # FIXME: Does not work in some cases, such as 'max()'
        return self._combine_predictions(rows_pred, cols_pred).reshape(-1)

    @available_if(_secondary_estimators_have("score"))
    def score(self, X, y=None):
        return type(self.secondary_rows_estimator_).score(self, X, y.reshape(-1))

    @property
    def _estimator_type(self):
        return getattr(self.secondary_rows_estimator, "_estimator_type", None)

    @property
    def classes_(self):
        """The classes labels. Only exist if the estimator is a classifier."""
        # FIXME: The secondary columns estimator generates different classes.
        return np.unique(self.secondary_rows_estimator_.classes_)

    def _more_tags(self):
        # check if the primary estimator expects pairwise input (the primary
        # columns is ensured to be pairwise by the _check_estimators method)
        # TODO: separate into pairwise_rows and pairwise_cols.
        return {"pairwise": _safe_tags(self.primary_rows_estimator, "pairwise")}

    @property
    def n_features_in_(self):
        """Number of features seen during fit."""
        return (
            self.primary_rows_estimator_.n_features_in_
            + self.primary_cols_estimator_.n_features_in_
        )

    @property
    def feature_names_in_(self):
        """Names of features seen during first step `fit` method."""
        return (
            self.primary_rows_estimator_.feature_names_in_
            + self.primary_cols_estimator_.feature_names_in_
        )

    def __sklearn_is_fitted__(self):
        """Indicate whether the primary estimators have been fit."""
        try:
            check_is_fitted(self.primary_rows_estimator_)
            check_is_fitted(self.primary_cols_estimator_)
            return True
        except NotFittedError:
            # Let the caller check_is_fitted() raise the error
            return False


class MultipartiteTransformerWrapper(
    BaseMultipartiteEstimator,
    TransformerMixin,
):
    """Manages a transformer for each feature space in multipartite datasets."""

    def __init__(
        self,
        transformers: BaseEstimator | Sequence[BaseEstimator],
        ndim: int | None = 2,
    ):
        self.transformers = transformers
        self.ndim = ndim

    def fit(self, X, y=None):
        self._set_transformers()
        self._validate_data(X, y)
        for Xi, yi, transformer in self._roll_axes(X, y):
            transformer.fit(Xi, yi)

    def transform(self, X, y=None):
        self._validate_data(X, y)
        return [transformer.transform(Xi) for Xi, _, transformer in self._roll_axes(X)]

    def fit_transform(self, X, y=None):
        self._set_transformers()
        self._validate_data(X, y)
        return [
            transformer.fit_transform(Xi, yi)
            for Xi, yi, transformer in self._roll_axes(X, y)
        ]

    def _set_transformers(self):
        """Sets self.transformers_ and self.ndim_ during fit."""
        if isinstance(self.transformers, Sequence):
            if self.ndim is None:
                self.ndim_ = len(self.transformers)

            elif self.ndim != len(self.transformers):
                raise ValueError(
                    "'self.ndim' must correspond to the number of"
                    " transformers in self.transformers"
                )
            else:
                self.ndim_ = self.ndim

            self.transformers_ = [clone(t) for t in self.transformers]

        else:  # If a single transformer provided
            if self.ndim is None:
                raise ValueError(
                    "'ndim' must be provided if 'transformers' is"
                    " a single transformer."
                )
            self.ndim_ = self.ndim
            self.transformers_ = [clone(self.transformers) for _ in range(self.ndim_)]

    def _validate_data(self, X, y):
        if len(X) != self.ndim_:
            raise ValueError(
                f"Wrong dimension. X has {len(X)} items, was exp" "ecting {self.ndim}."
            )
        # FIXME: validate
        # super()._validate_data(X, y, validate_separately=True)

    def _roll_axes(self, X, y=None):
        if not hasattr(self, "transformers_"):
            raise AttributeError(
                "One must call _set_transformers() before"
                "attempting to call _roll_axes()"
            )

        for ax in range(self.ndim_):
            yield X[ax], y, self.transformers_[ax]

            if y is not None:
                y = np.moveaxis(y, 0, -1)

    def __sklearn_is_fitted__(self):
        """Indicate whether the transformers have been fit."""
        if not hasattr(self, "ndim_"):
            return False
        try:
            map(check_is_fitted, self.transformers_)
        except NotFittedError:
            # Let the caller check_is_fitted() raise the error
            return False


# TODO: sampler wrapper and transformer wrapper could share code
class MultipartiteSamplerWrapper(BaseMultipartiteSampler):
    """Manages a sampler for each feature space in multipartite datasets."""

    def __init__(
        self,
        samplers: BaseEstimator | Sequence[BaseEstimator],
        ndim: int | None = 2,
    ):
        self.samplers = samplers
        self.ndim = ndim

    def _fit_resample(self, X, y):
        # NOTE: ressampled y is discarded!
        self._set_samplers()
        self._validate_data(X, y)
        return (
            [
                # FIXME: imblearn cannot deal with multi-column y
                # sampler.fit_resample(Xi, yi)[0]  # skip validation
                sampler._fit_resample(Xi, yi)[0]
                for Xi, yi, sampler in self._roll_axes(X, y)
            ],
            y,
        )

    def _set_samplers(self):
        """Sets self.samplers_ and self.ndim_ during fit."""
        if isinstance(self.samplers, Sequence):
            if self.ndim is None:
                self.ndim_ = len(self.samplers)

            elif self.ndim != len(self.samplers):
                raise ValueError(
                    "'self.ndim' must correspond to the number of"
                    " samplers in self.samplers"
                )
            else:
                self.ndim_ = self.ndim

            self.samplers_ = [clone(t) for t in self.samplers]

        else:  # If a single sampler provided
            if self.ndim is None:
                raise ValueError(
                    "'ndim' must be provided if 'samplers' is" " a single sampler."
                )
            self.ndim_ = self.ndim
            self.samplers_ = [clone(self.samplers) for _ in range(self.ndim_)]

    def _validate_data(self, X, y):
        if len(X) != self.ndim_:
            raise ValueError(
                f"Wrong dimension. X has {len(X)} items, was exp" "ecting {self.ndim}."
            )
        # FIXME: validate
        # super()._validate_data(X, y, validate_separately=True)

    def _roll_axes(self, X, y=None):
        if not hasattr(self, "samplers_"):
            raise AttributeError(
                "One must call _set_samplers() before" "attempting to call _roll_axes()"
            )

        for ax in range(self.ndim_):
            yield X[ax], y, self.samplers_[ax]

            if y is not None:
                y = np.moveaxis(y, 0, -1)



from sklearn.tree._tree cimport (
    DTYPE_t,   # Type of X
    DOUBLE_t,  # Type of y, sample_weight
    SIZE_t,    # Type for indices and counters
)
from ._axis_criterion cimport AxisCriterion


cdef class BaseBipartiteCriterion:
    """Abstract base class."""


cdef class BipartiteCriterion:
    """Abstract base class."""
    cdef const DTYPE_t[:, :] X_rows
    cdef const DTYPE_t[:, :] X_cols
    cdef const DOUBLE_t[:, :] y

    cdef const DOUBLE_t[:] row_weights
    cdef const DOUBLE_t[:] col_weights
    cdef const SIZE_t[:] row_indices
    cdef const SIZE_t[:] col_indices
    cdef SIZE_t[2] start
    cdef SIZE_t[2] end

    # NOTE: A source of confusion is that sometimes n_outputs is actually
    #       treated as the number of outputs, but sometimes it is just an
    #       alias for y.shape[1]. In monopartite data, they have the same
    #       value, but for bipartite interaction data one should have this
    #       distinction in mind.
    cdef SIZE_t n_outputs
    cdef SIZE_t n_outputs_rows
    cdef SIZE_t n_outputs_cols

    cdef SIZE_t n_rows
    cdef SIZE_t n_cols
    cdef SIZE_t n_node_rows
    cdef SIZE_t n_node_cols
    cdef double sq_sum_total
    cdef double[::1] sum_total
    cdef double weighted_n_rows
    cdef double weighted_n_cols
    cdef double weighted_n_samples

    cdef double weighted_n_node_rows
    cdef double weighted_n_node_cols
    cdef double weighted_n_node_samples

    cdef int init(
        self,
        const DTYPE_t[:, :] X_rows,
        const DTYPE_t[:, :] X_cols,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] row_weights,
        const DOUBLE_t[:] col_weights,
        double weighted_n_rows,
        double weighted_n_cols,
        const SIZE_t[:] row_indices,
        const SIZE_t[:] col_indices,
        SIZE_t[2] start,
        SIZE_t[2] end,
    ) nogil except -1

    cdef void node_value(self, double* dest) nogil

    cdef double node_impurity(self) nogil

    cdef void children_impurity(
            self,
            double* impurity_left,
            double* impurity_right,
            SIZE_t axis,
    ) nogil

    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double impurity_left,
        double impurity_right,
        SIZE_t axis,
    ) nogil


cdef class GMO(BipartiteCriterion):
    """Applies Predictive Bi-Clustering Trees method.

    See [Pliakos _et al._, 2018](https://doi.org/10.1007/s10994-018-5700-x).
    """
    cdef:
        SIZE_t max_n_classes
        public AxisCriterion criterion_rows
        public AxisCriterion criterion_cols

        void* _get_criterion(self, SIZE_t axis) nogil



from sklearn.utils._tags import _safe_tags
from imblearn.pipeline import make_pipeline
from .wrappers import (
    MultipartiteTransformerWrapper,
    MultipartiteSamplerWrapper,
)


# TODO: test get/set_params
def make_multipartite_pipeline(*steps, ndim=2, memory=None, verbose=False):
    """Utility function to create pipelines for multipartite data.

    It wraps monopartite transformers with MultipartiteTransformerWrapper.
    """
    # Make pipeline before modifying to preserve step names
    pipe = make_pipeline(*steps, memory=memory, verbose=verbose)

    # FIXME: what if {"multipartite": False}?
    if "multipartite" not in _safe_tags(pipe.steps[-1][1]):
        raise ValueError(
            "Last step of a multipartite pipeline must be a multipartite "
            f"estimator, not {pipe.steps[-1][1]}."
        )

    for i, step in enumerate(pipe.steps[:-1]):
        # FIXME: what if {"multipartite": False}?
        if "multipartite" not in _safe_tags(step[1]):
            estimator = step[1]
            if (
                hasattr(estimator, "_estimator_type")
                and estimator._estimator_type == "sampler"
            ):
                wrapped = MultipartiteSamplerWrapper(estimator, ndim=ndim)
            elif hasattr(estimator, "transform"):
                wrapped = MultipartiteTransformerWrapper(estimator, ndim=ndim)
            else:
                raise ValueError(
                    f"Pipeline step of type {type(estimator)} is not a sampler"
                    " or transformer."
                )

            pipe.steps[i] = (step[0], wrapped)
    
    return pipe



import pytest
import numpy as np
from sklearn.datasets import make_regression
from sklearn.utils._testing import assert_allclose
from sklearn.metrics import mean_squared_error
from sklearn.metrics.pairwise import rbf_kernel, euclidean_distances
from bipartite_learn.neighbors import WeightedNeighborsRegressor


@pytest.fixture
def data():
    X, y = make_regression(n_samples=1000, n_features=500, n_targets=100)
    X.setflags(write=False)
    return X, y


def test_weighted_neighbors_euclidean(data):
    X, y = data

    rng = np.random.default_rng()
    X_test = rng.random((X.shape[0] // 2, X.shape[1]))

    wnr = WeightedNeighborsRegressor()
    wnr.fit(X, y)

    assert_allclose(wnr.predict(X), y)

    y_pred = wnr.predict(X_test)
    euclidean = np.sqrt(((X_test[:, np.newaxis, :] - X) ** 2).sum(axis=-1))
    w = 1 / euclidean
    w /= w.sum(axis=0, keepdims=True)
    assert w.sum(axis=0, keepdims=True).shape == (1, X.shape[0])

    assert_allclose(y_pred, (w @ y))


def test_weighted_neighbors_precomputed(data):
    X, y = data
    X = euclidean_distances(X) + 1e-6  # to avoid zero distances

    wnr = WeightedNeighborsRegressor(metric="precomputed")
    wnr.fit(X, y)
    wnr.predict(X[: X.shape[0] // 2])

    y_pred = wnr.predict(X)

    w = 1 / X
    w /= w.sum(axis=1, keepdims=True)
    assert (y_pred == w @ y).all()


def test_weighted_neighbors_precomputed_similarity(data):
    X, y = data
    X = rbf_kernel(X)

    wnr = WeightedNeighborsRegressor(
        metric="precomputed",
        weights="similarity",
    )
    wnr.fit(X, y)
    y_pred = wnr.predict(X)

    w = X / X.sum(axis=1, keepdims=True)

    assert np.allclose(y_pred, (w @ y), rtol=1e-5, atol=1e-5)


def test_weighted_neighbors_regressor_validate_params(data):
    X, y = data
    X = rbf_kernel(X)
    wnr = WeightedNeighborsRegressor(metric="euclidean", weights="similarity")

    with pytest.raises(
        ValueError,
        match="weights='similarity' is not supported for string-valued.*",
    ):
        wnr.fit(X, y)



import pytest
import logging
from copy import deepcopy
from pathlib import Path
import numpy as np
from bipartite_learn import datasets
from bipartite_learn.datasets.loader import (
    BaseFileLoader,
    BaseRemoteFileLoader,
    BipartiteDatasetLoader,
    InvalidChecksumError,
)


class BasicTextLoader(BaseFileLoader):
    def load(self, as_frame=False):
        with self.local_path.open() as f:
            return f.read()


class BasicRemoteTextLoader(BaseRemoteFileLoader):
    def download(self) -> Path:
        logging.info(f"Downloading {self.url} to {self.local_path}")
        return super().download()

    def load_local(self, as_frame=False):
        with self.local_path.open() as f:
            return f.read()


@pytest.fixture(scope="module")
def remote_url():
    return (
        "https://raw.githubusercontent.com/pedroilidio/bipartite_learn"
        "/main/README.md"
    )


@pytest.fixture(scope="module")
def checksum():
    return (
        "395fcfcdb437a918915f35bb7e6b1f7ef2a4dc10ef980340f1949ebfaac3112b"
    )


@pytest.fixture
def local_text_loader(tmp_path):
    out = tmp_path / "test.txt"
    with out.open("w") as f:
        f.write("test content")

    return BasicTextLoader(filepath=out)


@pytest.fixture
def remote_text_loader(tmp_path, remote_url):
    return BasicRemoteTextLoader(
        url=remote_url,
        base_dir=tmp_path,
    )


@pytest.fixture
def bipartite_dataset_loader(tmp_path, local_text_loader, remote_text_loader):
    return BipartiteDatasetLoader(
        X_loader=[local_text_loader, remote_text_loader],
        y_loader=remote_text_loader,
        base_dir=tmp_path,
        filepath="test_dataset",
    )


class TestLocalFileLoader:
    def test_local_path(self, tmp_path):
        loader = BasicTextLoader(filepath=tmp_path)
        assert loader.local_path == tmp_path.resolve()
    
    def test_set_description_str(self, local_text_loader):
        description = "test description"
        local_text_loader.set_description(description)
        assert local_text_loader.description == description
        assert local_text_loader.get_description() == description

    def test_set_description_loader(self, local_text_loader):
        local_text_loader.set_description(local_text_loader)
        assert local_text_loader.description is local_text_loader
        assert local_text_loader.get_description() == "test content"

    def test_set_description_remote_loader(
        self, local_text_loader, remote_text_loader,
    ):
        local_text_loader.set_description(remote_text_loader)
        assert local_text_loader.description is remote_text_loader
        assert (
            local_text_loader.get_description()
            == remote_text_loader.load_local()
        )

    def test_set_description_type_error(self, local_text_loader):
        with pytest.raises(TypeError):
            local_text_loader.set_description(123)
    

class TestRemoteFileLoader:
    def test_infer_filepath(self):
        url = (
            "https://archive.ics.uci.edu/ml/machine-learning-databases/iris"
            "/iris.data?example=1"
        )
        loader = BasicRemoteTextLoader(url=url)
        assert loader.filepath == Path("iris.data")

    def test_set_description_loader(self, remote_text_loader, local_text_loader):
        remote_text_loader.set_description(local_text_loader)
        assert remote_text_loader.description is local_text_loader
        assert remote_text_loader.get_description() == "test content"

    def test_set_description_remote_loader(self, remote_text_loader):
        main_loader = deepcopy(remote_text_loader)
        main_loader.set_description(remote_text_loader)
        assert (
            main_loader.get_description() == remote_text_loader.load_local()
        )
        assert remote_text_loader.base_dir == main_loader.base_dir

    def test_set_base_dir(self, tmp_path, remote_text_loader):
        remote_text_loader.set_base_dir(tmp_path)
        assert remote_text_loader.base_dir == tmp_path

    def test_local_path(self, tmp_path, remote_text_loader):
        filepath = remote_text_loader.filepath
        remote_text_loader.set_base_dir(tmp_path)
        remote_text_loader.local_path == (tmp_path / filepath).resolve()

    def test_rebase_dir(self, tmp_path, remote_text_loader):
        loader = remote_text_loader
        filepath = remote_text_loader.filepath
        base_dir = tmp_path / "test/test1/test2"

        loader.set_base_dir("test2")
        loader.rebase_dir("test1")
        loader.rebase_dir("test")
        loader.rebase_dir(tmp_path)

        assert loader.base_dir == base_dir
        assert loader.local_path == base_dir / filepath
    
    def test_download(self, remote_text_loader):
        remote_text_loader.download()
        assert remote_text_loader.local_path.exists()

    def test_checksum_fail(self, remote_text_loader):
        remote_text_loader.checksum = "123"
        with pytest.raises(InvalidChecksumError):
            remote_text_loader.download()

    def test_correct_checksum(self, remote_text_loader, checksum):
        remote_text_loader.checksum = checksum
        remote_text_loader.download()
        assert remote_text_loader.local_path.exists()

    def test_clear_local(self, remote_text_loader):
        remote_text_loader.download()
        assert remote_text_loader.local_path.exists()
        remote_text_loader.clear_local()
        assert not remote_text_loader.local_path.exists()


class TestBipartiteDatasetLoader:
    def test_set_base_dir(self, bipartite_dataset_loader, tmp_path):
        loader = bipartite_dataset_loader
        local_path = tmp_path / loader.filepath 
        loader.set_base_dir(tmp_path)

        assert loader.base_dir == tmp_path
        assert loader.local_path == local_path
        assert loader.X_loader[1].base_dir == local_path
        assert loader.y_loader.base_dir == local_path

    def test_download(self, bipartite_dataset_loader, checksum):
        loader = bipartite_dataset_loader
        loader.X_loader[1].checksum = checksum
        loader.download()

        assert loader.X_loader[0].local_path.exists()
        assert loader.X_loader[1].local_path.exists()
        assert loader.y_loader.local_path.exists()

    def test_clear_local(self, bipartite_dataset_loader, checksum):
        loader = bipartite_dataset_loader
        loader.X_loader[1].checksum = checksum

        loader.download()
        assert loader.y_loader.local_path.exists()
        loader.clear_local()

        assert loader.X_loader[0].local_path.exists()  # Loader is local
        assert not loader.X_loader[1].local_path.exists()
        assert not loader.y_loader.local_path.exists()
        assert not loader.local_path.exists()
    
    def test_load(self, bipartite_dataset_loader):
        loader = bipartite_dataset_loader
        X, y = loader.load()

        assert X[0] == loader.X_loader[0].load()
        assert X[1] == loader.X_loader[1].load()
        assert y == loader.y_loader.load()


@pytest.mark.parametrize(
    "loader_class", [
        datasets.EnzymesLoader,
        datasets.IonChannelsLoader,
        datasets.GPCRLoader,
        datasets.NuclearReceptorsLoader,
    ],
    ids=[
        "enzymes",
        "ion_channels",
        "gpcr",
        "nuclear_receptors",
    ],
)
def test_all_dataset_loaders(loader_class, tmp_path):
    loader = loader_class(tmp_path)
    X, y = loader.load(as_frame=False)
    assert isinstance(X, list)
    assert isinstance(y, np.ndarray)
    assert isinstance(X[0], np.ndarray)
    assert isinstance(X[1], np.ndarray)
    assert X[0].shape[0] == y.shape[0]
    assert X[1].shape[0] == y.shape[1]


from sklearn.tree._splitter cimport SplitRecord, Splitter

from sklearn.tree._tree cimport DTYPE_t          # Type of X
from sklearn.tree._tree cimport DOUBLE_t         # Type of y, sample_weight
from sklearn.tree._tree cimport SIZE_t           # Type for indices and counters

from ._bipartite_criterion cimport BipartiteCriterion


# Data to track sample split
cdef struct MultipartiteSplitRecord:
    SplitRecord split_record  # Monopartite object to use struct composition.
    SIZE_t axis               # Axis in which the split occurred.


cdef class BipartiteSplitter:
    # Wrapper class to coordinate 2 Splitters at the same time.

    # The splitter searches in the input space for a feature and a threshold
    # to split the samples samples[start:end].
    #
    # The impurity computations are delegated to a criterion object.
    cdef public Splitter splitter_rows
    cdef public Splitter splitter_cols
    cdef public BipartiteCriterion bipartite_criterion
    cdef public SIZE_t n_row_features  # Number of row features (X[0].shape[0]).
    # TODO: X dtype must be DOUBLE_t, not DTYPE_t (float32) to use
    # semisupervision.
    cdef const DTYPE_t[:, :] X_rows
    cdef const DTYPE_t[:, :] X_cols
    cdef const DOUBLE_t[:, :] y

    cdef SIZE_t n_rows
    cdef SIZE_t n_cols
    cdef SIZE_t n_samples
    cdef double weighted_n_rows
    cdef double weighted_n_cols
    cdef double weighted_n_samples

    cdef public SIZE_t min_samples_leaf
    cdef public SIZE_t min_rows_leaf
    cdef public SIZE_t min_cols_leaf
    cdef public double min_weight_leaf
    # TODO
    # cdef public double min_row_weight_leaf
    # cdef public double min_col_weight_leaf

    cdef const DOUBLE_t[:] row_weights
    cdef const DOUBLE_t[:] col_weights
    cdef SIZE_t[::1] row_indices  # TODO: Drop contiguity?
    cdef SIZE_t[::1] col_indices
    cdef SIZE_t[2] start
    cdef SIZE_t[2] end

    # Methods
    cdef int init(
        self,
        object X,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight,
    ) except -1

    cdef int node_reset(
        self,
        SIZE_t[2] start,
        SIZE_t[2] end,
        double* weighted_n_node_samples
    ) nogil except -1

    cdef int node_split(
        self,
        double impurity,
        MultipartiteSplitRecord* split,
        SIZE_t[2] n_constant_features,
    ) nogil except -1

    cdef void node_value(self, double* dest) nogil

    cdef double node_impurity(self) nogil


from ._axis_criterion cimport BaseComposableCriterion
from sklearn.tree._criterion cimport Criterion

cdef class BaseUnsupervisedCriterion(BaseComposableCriterion):
    pass

cdef class UnsupervisedWrapperCriterion(BaseUnsupervisedCriterion):
    cdef Criterion criterion
    cdef inline void _copy_node_wise_attributes(self) noexcept nogil
    cdef inline void _copy_position_wise_attributes(self) noexcept nogil

cdef class PairwiseCriterion(UnsupervisedWrapperCriterion):
    pass

cdef class BaseKernelCriterion(BaseUnsupervisedCriterion):
    cdef double sum_total
    cdef double sum_left
    cdef double sum_right


"""Distance Weighted Neighbors Regression."""

# Author: Pedro Ilídio <ilidio@alumni.usp.br>
# Adapted from scikit-learn.

# License: BSD 3 clause (C) Pedro Ilídio

import numpy as np

from joblib import effective_n_jobs
from sklearn.neighbors._base import (
    _get_weights, _check_precomputed, NeighborsBase, KNeighborsMixin,
)
from sklearn.base import RegressorMixin, _fit_context
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.utils._param_validation import StrOptions


class WeightedNeighborsRegressor(KNeighborsMixin, RegressorMixin, NeighborsBase):
    f"""Regression based on distance-weighted neighbors.
    The target is predicted by distance-weighted interpolation of the targets
    associated with the instances in the training set. Corresponds to
    `KneighborsRegressor(n_neighbors=X_train.shape[0])`, but much faster.
    Parameters
    ----------
    weights : {'distance', 'similarity'}, callable or None, default='distance'
        Weight function used in prediction.  Possible values:
        - 'distance' : weight points by the inverse of their distance.
          in this case, closer neighbors of a query point will have a
          greater influence than neighbors which are further away.
        - 'similarity' : the same as 'distance', but assumes X or self.metric(X)
          is a similarity matrix rather than a distance matrix (for self.metric
          == 'precomputed' and callable self.metric, respectively).
        - [callable] : a user-defined function which accepts an
          array of distances, and returns an array of the same shape
          containing the weights.
        Distance weights are used by default.
    p : int, default=2
        Power parameter for the Minkowski metric. When p = 1, this is
        equivalent to using manhattan_distance (l1), and euclidean_distance
        (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used.
    metric : str or callable, default='minkowski'
        Metric to use for distance computation. Default is "minkowski", which
        results in the standard Euclidean distance when p = 2. See the
        documentation of `scipy.spatial.distance
        <https://docs.scipy.org/doc/scipy/reference/spatial.distance.html>`_ and
        the metrics listed in
        :class:`~sklearn.metrics.pairwise.distance_metrics` for valid metric
        values.
        If metric is "precomputed", X is assumed to be a kernel/distance matrix
        and must be square during fit. X may be a :term:`sparse graph`, in which
        case only "nonzero" elements may be considered neighbors.
        If metric is a callable function, it takes two arrays representing 1D
        vectors as inputs and must return one value indicating the distance
        between those vectors. This works for Scipy's metrics, but is less
        efficient than passing the metric name as a string.
    metric_params : dict, default=None
        Additional keyword arguments for the metric function.
    n_jobs : int, default=None
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
        Doesn't affect :meth:`fit` method.
    Attributes
    ----------
    effective_metric_ : str or callable
        The distance metric to use. It will be same as the `metric` parameter
        or a synonym of it, e.g. 'euclidean' if the `metric` parameter set to
        'minkowski' and `p` parameter set to 2.
    effective_metric_params_ : dict
        Additional keyword arguments for the metric function. For most metrics
        will be same with `metric_params` parameter, but may also contain the
        `p` parameter value if the `effective_metric_` attribute is set to
        'minkowski'.
    n_features_in_ : int
        Number of features seen during :term:`fit`.
        .. versionadded:: 0.24
    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.
        .. versionadded:: 1.0
    n_samples_fit_ : int
        Number of samples in the fitted data.
    See Also
    --------
    NearestNeighbors : Unsupervised learner for implementing neighbor searches.
    RadiusNeighborsRegressor : Regression based on neighbors within a fixed radius.
    KNeighborsRegressor : Classifier implementing the k-nearest neighbors vote.
    KNeighborsClassifier : Classifier implementing the k-nearest neighbors vote.
    RadiusNeighborsClassifier : Classifier implementing
        a vote among neighbors within a given radius.
    Examples
    --------
    >>> X = [[0], [1], [2], [3]]
    >>> y = [0, 0, 1, 1]
    >>> from sklearn.neighbors import KNeighborsRegressor
    >>> neigh = KNeighborsRegressor(n_neighbors=2)
    >>> neigh.fit(X, y)
    KNeighborsRegressor(...)
    >>> print(neigh.predict([[1.5]]))
    [0.5]
    """

    _parameter_constraints: dict = {
        **NeighborsBase._parameter_constraints,
        "weights": [StrOptions({"distance", "similarity"}), callable, None],
    }
    _parameter_constraints.pop("radius")
    _parameter_constraints.pop("n_neighbors")
    _parameter_constraints.pop("leaf_size")
    _parameter_constraints.pop("algorithm")

    def __init__(
        self,
        *,
        weights="distance",
        p=2,
        metric="minkowski",
        metric_params=None,
        n_jobs=None,
    ):
        super().__init__(
            n_neighbors=None,
            algorithm="brute",
            leaf_size=None,
            metric=metric,
            p=p,
            metric_params=metric_params,
            n_jobs=n_jobs,
        )
        self.weights = weights

    def _more_tags(self):
        # For cross-validation routines to split data correctly
        return {"pairwise": self.metric == "precomputed"}

    def _validate_params(self):
        if (
            self.weights == "similarity"
            and self.metric != "precomputed"
            and isinstance(self.metric, str)
        ):
            raise ValueError(
                "weights='similarity' is not supported for string-valued"
                f" metrics other than 'precomputed' ({self.metric=!r})",
            )
        super()._validate_params()

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y):
        """Fit the k-nearest neighbors regressor from the training dataset.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features) or \
                (n_samples, n_samples) if metric='precomputed'
            Training data.
        y : {array-like, sparse matrix} of shape (n_samples,) or \
                (n_samples, n_outputs)
            Target values.
        Returns
        -------
        self : KNeighborsRegressor
            The fitted k-nearest neighbors regressor.
        """
        self.n_neighbors = X.shape[0]  # TODO: optimize other methods that use it
        return self._fit(X, y)

    def predict(self, X):
        """Predict the target for the provided data.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_queries, n_features), \
                or (n_queries, n_indexed) if metric == 'precomputed'
            Test samples.
        Returns
        -------
        y : ndarray of shape (n_queries,) or (n_queries, n_outputs), dtype=int
            Target values.
        """
        if self.metric == "precomputed":
            dist = _check_precomputed(X)
        else:
            X = self._validate_data(
                X, accept_sparse="csr", reset=False, order="C",
            )

            dist = pairwise_distances(
                X,
                self._fit_X,
                metric=self.effective_metric_,
                n_jobs=effective_n_jobs(self.n_jobs),
                **self.effective_metric_params_,
            )

        if self.weights == "similarity":
            weights = dist
        else:
            # Default weights is "distance", no more "uniform"
            weights = _get_weights(dist, self.weights or "distance")

        # Normalize the weights to sum up to 1 on each row:
        #   The weights of the sum over the training samples, one weight for each
        #   sample, must be 1 in total. Since y_pred = weights @ y, each column of
        #   weights represents a training instance.
        weights /= np.sum(weights, axis=1, keepdims=True)

        _y = self._y
        if _y.ndim == 1:
            _y = _y.reshape((-1, 1))

        y_pred = weights @ _y

        if self._y.ndim == 1:
            y_pred = y_pred.ravel()

        return y_pred



# cython: boundscheck=False

# Author: Pedro Ilídio <pedrilidio@gmail.com>
# Modified from scikit-learn.
#
# License: BSD 3 clause

from cython.view cimport array as cvarray
from libc.stdlib cimport malloc, free
from libc.string cimport memset, memcpy
from libc.stdint cimport SIZE_MAX
cimport numpy as cnp
import numpy as np
from sklearn.tree._utils cimport log

cnp.import_array()
cdef double NAN = np.nan


cdef class BaseComposableCriterion(Criterion):
    cdef double _proxy_improvement_factor(self) noexcept nogil:
        """If improvement = proxy_improvement / a + b, this method returns a.

        This is useful when defining proxy impurity improvements for
        compositions of Criterion objects.
        """


cdef class AxisCriterion(BaseComposableCriterion):
    """Criterion that is able to select a subset of columns to consider.

    Criterion objects gather methods to evaluate the impurity of a partition
    and the impurity improvement of a split, each subclass implementing a
    different impurity metric.

    They are held by a Splitter object that provides the y target matrix using
    Criterion.init() (called in Splitter.node_reset()) and directs the
    Criterion using Criterion.update() to specific split positions where the
    children impurities and impurity improvement can be calculated.

    AxisCriterion is analogous to sklearn.tree._criterion.Criterion, but can
    also calculate the impurity on the other axis for the current node and
    children (as if each child partition were transposed), and has an
    additional init_columns() method to receive column indices and column
    weights.

    y's contiguity is also dismissed, since y columns will now be randomly
    acessed.
    """
    def __getstate__(self):
        return {}

    def __setstate__(self, d):
        pass

    def __cinit__(self):
        self.y = None
        self.y_ = None
        self._columns_are_set = False
        self._cached_pos = SIZE_MAX
        self._cached_rows_impurity_left = 0.0
        self._cached_rows_impurity_right = 0.0
        self._cached_cols_impurity_left = 0.0
        self._cached_cols_impurity_right = 0.0
        # TODO: use cached node impurity
        self._cached_rows_node_impurity = 0.0
        self._cached_cols_node_impurity = 0.0

        self.col_indices = None
        # TODO: Possibly remove if Splitter can find split without reordering
        # sample_indices. The upside is that we reduce memory usage, the
        # downside is that we lose C contiguity, which seems important here.
        self._node_col_indices = None

    cdef int init(
        self,
        const DOUBLE_t[:, ::1] y,
        const DOUBLE_t[:] sample_weight,
        double weighted_n_samples,
        const SIZE_t[:] sample_indices,
        SIZE_t start,
        SIZE_t end,
    ) except -1 nogil:
        with gil:
            raise RuntimeError("Please use axis_init() instead of init()")
    
    cdef void init_columns(
        self,
        const DOUBLE_t[:] col_weights,
        double weighted_n_cols,
        SIZE_t[:] col_indices,
        SIZE_t start_col,
        SIZE_t end_col,
    ) nogil:
        """Initialize the column sample indices and column weights.
        Parameters
        ----------
        col_weights : ndarray, dtype=DOUBLE_t
            The weight of each sample stored as a Cython memoryview.
        weighted_n_cols : double
            The total column weight of the y partition being considered
        col_indices : ndarray, dtype=SIZE_t
            A mask on the sample_indices. Indices of the sample_indices in X and y we want to use,
            where sample_indices[start:end] correspond to the sample_indices in this node.
        start : SIZE_t
            The first sample to be used on this node
        end : SIZE_t
            The last sample used on this node
        """
        self.col_indices = col_indices
        self.col_weights = col_weights
        self.start_col = start_col
        self.end_col = end_col

        self.n_node_cols = end_col - start_col
        self.weighted_n_cols = weighted_n_cols
        self.weighted_n_node_cols = 0.0

        cdef SIZE_t j, q
        cdef DOUBLE_t w = 1.0

        for q in range(start_col, end_col):
            j = col_indices[q]
            self._node_col_indices[q - start_col] = j

            if col_weights is not None:
                w = col_weights[j]
            self.weighted_n_node_cols += w

        self._columns_are_set = True
 
    cdef int axis_init(
        self,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight,
        const DOUBLE_t[:] col_weights,
        const SIZE_t[:] sample_indices,
        const SIZE_t[:] col_indices,
        double weighted_n_samples,
        double weighted_n_cols,
        SIZE_t start,
        SIZE_t end,
        SIZE_t start_col,
        SIZE_t end_col,
    ) except -1 nogil:
        """Placeholder for a method which will initialize the criterion.
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.
        Parameters
        ----------
        y : ndarray, dtype=DOUBLE_t
            y is a buffer that can store values for n_outputs target variables
            stored as a Cython memoryview.
        sample_weight : ndarray, dtype=DOUBLE_t
            The weight of each sample stored as a Cython memoryview.
        weighted_n_samples : double
            The total weight of the sample_indices being considered
        sample_indices : ndarray, dtype=SIZE_t
            A mask on the sample_indices. Indices of the sample_indices in X and y we want to use,
            where sample_indices[start:end] correspond to the sample_indices in this node.
        start : SIZE_t
            The first sample to be used on this node
        end : SIZE_t
            The last sample used on this node
        """
        pass
   
    cdef double node_impurity(self) noexcept nogil:
        """Evaluate the impurity of the current node.
        Evaluate the MSE criterion as impurity of the current node,
        i.e. the impurity of sample_indices[start:end]. The smaller the impurity the
        better.
        """
        cdef:
            SIZE_t k
            double rows_impurity
            double cols_impurity

        self.node_axes_impurities(&rows_impurity, &cols_impurity)

        return 0.5 * (rows_impurity + cols_impurity)

    cdef void children_impurity(
        self,
        double* impurity_left,
        double* impurity_right
    ) noexcept nogil:
        """Evaluate the impurity in children nodes.
        i.e. the impurity of the left child (sample_indices[start:pos]) and the
        impurity the right child (sample_indices[pos:end]).
        """
        cdef:
            double rows_impurity_left
            double rows_impurity_right
            double cols_impurity_left
            double cols_impurity_right

        self.children_axes_impurities(
            &rows_impurity_left,
            &rows_impurity_right,
            &cols_impurity_left,
            &cols_impurity_right,
        )

        impurity_left[0] = 0.5 * (rows_impurity_left + cols_impurity_left)
        impurity_right[0] = 0.5 * (rows_impurity_right + cols_impurity_right)
 
    cdef void node_axes_impurities(
        self,
        double* rows_impurity,
        double* cols_impurity,
    ) noexcept nogil:
        """Evaluate the impurity of the current node in both directions.
        Evaluate the MSE criterion as impurity of the current node,
        i.e. the impurity of sample_indices[start:end]. The smaller the impurity the
        better.
        """

    cdef void children_axes_impurities(
        self,
        double* rows_impurity_left,
        double* rows_impurity_right,
        double* cols_impurity_left,
        double* cols_impurity_right,
    ) noexcept nogil:
        """Evaluate the impurity in children nodes, in both directions.
        i.e. the impurity of the left child (sample_indices[start:pos]) and the
        impurity the right child (sample_indices[pos:end]).
        Should employ self._cached_* attributes to minimize computation.
        """

    cdef double proxy_impurity_improvement(self) noexcept nogil:
        """Compute a proxy of the impurity reduction.

        This method is used to speed up the search for the best split.
        It is a proxy quantity such that the split that maximizes this value
        also maximizes the impurity improvement. It neglects all constant terms
        of the impurity decrease for a given split.

        The absolute impurity improvement is only computed by the
        impurity_improvement method once the best split has been found.
        """
        cdef:
            double rows_impurity_left
            double rows_impurity_right
            double cols_impurity_left   # Discarded
            double cols_impurity_right  # Discarded

        self.children_axes_impurities(
            &rows_impurity_left,
            &rows_impurity_right,
            &cols_impurity_left,
            &cols_impurity_right,
        )

        return (- self.weighted_n_right * rows_impurity_right
                - self.weighted_n_left * rows_impurity_left)

    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double impurity_left,
        double impurity_right,
    ) noexcept nogil:
        """Compute the improvement in impurity.

        Differently from the usual sklearn objects, the impurity improvement
        is NOT:
            (imp_parent - imp_children) / imp_parent
        but instead:
            0.5 * (
                (rows_imp_parent - rows_imp_children) / rows_imp_parent
                +
                (cols_imp_parent - cols_imp_children) / cols_imp_parent
            )
        In many cases, however, columns impurity improvement is zero, and we
        return only the impurity over rows.
        """
        return (
            (self.weighted_n_node_samples / self.weighted_n_samples)
            * (self.weighted_n_node_cols / self.weighted_n_cols)
            * (
                impurity_parent
                - impurity_right
                    * (self.weighted_n_right / self.weighted_n_node_samples)
                - impurity_left
                    * (self.weighted_n_left / self.weighted_n_node_samples)
            )
        )

    cdef void node_value(self, double* dest) noexcept nogil:
        pass

    cdef void total_node_value(self, double* dest) noexcept nogil:
        """Compute a single node value for all targets, disregarding y's shape.

        This method is used instead of node_value() in cases where the
        different columns of y are *not* considered as different outputs, being
        usually equivalent to node_value if y were to be flattened, i.e.

            total_node_value(y) == node_value(y.reshape[-1, 1])

        Parameters
        ----------
        dest : double pointer
            The memory address which we will save the node value into.
        """
        pass
    
    cdef double _proxy_improvement_factor(self) noexcept nogil:
        """If improvement = proxy_improvement / a + b, this method returns a.

        This is useful when defining proxy impurity improvements for
        compositions of Criterion objects.
        """
        # The 2 factor accounts for the fact that the current axis' impurity is
        # averaged with the impurity on the other axis impurity, but the other
        # axis improvement will usually be zero.
        return (
            2.0
            * self.weighted_n_cols
            / self.weighted_n_node_cols
            * self.weighted_n_samples
        )


cdef class AxisRegressionCriterion(AxisCriterion):
    r"""Abstract regression criterion.
    This handles cases where the target is a continuous value, and is
    evaluated by computing the variance of the target values left and right
    of the split point. The computation takes linear time with `n_samples`
    by using ::
        var = \sum_i^n (y_i - y_bar) ** 2
            = (\sum_i^n y_i ** 2) - n_samples * y_bar ** 2
    """

    def __cinit__(self, SIZE_t n_outputs, SIZE_t n_samples):
        """Initialize parameters for this criterion.
        Parameters
        ----------
        n_outputs : SIZE_t
            The number of targets to be predicted
        n_samples : SIZE_t
            The total number of rows to fit on
        """
        # Default values
        self.start = 0
        self.pos = 0
        self.end = 0

        self.n_outputs = n_outputs
        self.n_samples = n_samples
        self.n_node_samples = 0
        self.weighted_n_node_samples = 0.0
        self.weighted_n_left = 0.0
        self.weighted_n_right = 0.0

        self.sq_sum_total = 0.0
        self.sum_total = np.zeros(n_outputs, dtype=np.float64)
        self.sum_left = np.zeros(n_outputs, dtype=np.float64)
        self.sum_right = np.zeros(n_outputs, dtype=np.float64)

        self._node_col_indices = np.empty(n_outputs, dtype=np.intp, order='C')

    def __reduce__(self):
        return (type(self), (self.n_outputs, self.n_samples), self.__getstate__())

    cdef int axis_init(
        self,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight,
        const DOUBLE_t[:] col_weights,
        const SIZE_t[:] sample_indices,
        const SIZE_t[:] col_indices,
        double weighted_n_samples,
        double weighted_n_cols,
        SIZE_t start,
        SIZE_t end,
        SIZE_t start_col,
        SIZE_t end_col,
    ) except -1 nogil:
        """Initialize the criterion.
        This initializes the criterion at node sample_indices[start:end] and children
        sample_indices[start:start] and sample_indices[start:end].
        """
        self._cached_pos = SIZE_MAX  # Important to reset cached values.
        self.init_columns(
            col_weights=col_weights,
            weighted_n_cols=weighted_n_cols,
            col_indices=col_indices,
            start_col=start_col,
            end_col=end_col,
        )
        # Initialize fields
        self.y_ = y
        self.sample_weight = sample_weight
        self.sample_indices = sample_indices
        self.start = start
        self.end = end
        self.n_node_samples = end - start
        self.weighted_n_samples = weighted_n_samples

        cdef:
            SIZE_t[::1] _node_col_indices = self._node_col_indices
            SIZE_t n_node_cols = self.n_node_cols
            SIZE_t i, j, p, k
            DOUBLE_t y_ij
            DOUBLE_t wi = 1.0, wj = 1.0
            DOUBLE_t row_sum

        self.sq_sum_total = 0.0
        self.sum_sq_row_sums = 0.0
        self.weighted_n_node_samples = 0.0
        memset(&self.sum_total[0], 0, self.n_node_cols * sizeof(double))

        for p in range(start, end):
            i = sample_indices[p]
            if sample_weight is not None:
                wi = sample_weight[i]

            self.weighted_n_node_samples += wi
            row_sum = 0.0

            for k in range(n_node_cols):
                j = _node_col_indices[k]
                if col_weights is not None:
                    wj = col_weights[j]

                y_ij = y[i, j]

                row_sum += wj * y_ij
                self.sum_total[k] += wi * y_ij
                self.sq_sum_total += wi * wj * y_ij * y_ij

            self.sum_sq_row_sums += wi * row_sum * row_sum

        # Reset to pos=start
        self.reset()
        return 0

    cdef int reset(self) except -1 nogil:
        """Reset the criterion at pos=start."""
        cdef SIZE_t n_bytes = self.n_node_cols * sizeof(double)
        memset(&self.sum_left[0], 0, n_bytes)
        memcpy(&self.sum_right[0], &self.sum_total[0], n_bytes)

        self.weighted_n_left = 0.0
        self.weighted_n_right = self.weighted_n_node_samples
        self.pos = self.start
        return 0

    cdef int reverse_reset(self) except -1 nogil:
        """Reset the criterion at pos=end."""
        cdef SIZE_t n_bytes = self.n_node_cols * sizeof(double)
        memset(&self.sum_right[0], 0, n_bytes)
        memcpy(&self.sum_left[0], &self.sum_total[0], n_bytes)

        self.weighted_n_right = 0.0
        self.weighted_n_left = self.weighted_n_node_samples
        self.pos = self.end
        return 0

    cdef int update(self, SIZE_t new_pos) except -1 nogil:
        """Updated statistics by moving sample_indices[pos:new_pos] to the left."""
        cdef:
            const DOUBLE_t[:] sample_weight = self.sample_weight
            const SIZE_t[:] sample_indices = self.sample_indices
            SIZE_t pos = self.pos
            SIZE_t end = self.end
            SIZE_t n_node_cols = self.n_node_cols

            SIZE_t[::1] _node_col_indices = self._node_col_indices

            SIZE_t i, p, k
            DOUBLE_t wi = 1.0

        # Update statistics up to new_pos
        #
        # Given that
        #           sum_left[x] +  sum_right[x] = sum_total[x]
        # and that sum_total is known, we are going to update
        # sum_left from the direction that require the least amount
        # of computations, i.e. from pos to new_pos or from end to new_pos.
        if (new_pos - pos) <= (end - new_pos):
            for p in range(pos, new_pos):
                i = sample_indices[p]
                if sample_weight is not None:
                    wi = sample_weight[i]

                for k in range(n_node_cols):
                    self.sum_left[k] += wi * self.y_[i, _node_col_indices[k]]

                self.weighted_n_left += wi
        else:
            self.reverse_reset()

            for p in range(end - 1, new_pos - 1, -1):
                i = sample_indices[p]
                if sample_weight is not None:
                    wi = sample_weight[i]

                for k in range(n_node_cols):
                    self.sum_left[k] -= wi * self.y_[i, _node_col_indices[k]]

                self.weighted_n_left -= wi

        self.weighted_n_right = (
            self.weighted_n_node_samples - self.weighted_n_left
        )

        for k in range(n_node_cols):
            self.sum_right[k] = self.sum_total[k] - self.sum_left[k]

        self.pos = new_pos
        return 0

    cdef void node_value(self, double* dest) noexcept nogil:
        """Compute the node value of sample_indices[start:end] into dest."""
        cdef SIZE_t j, k

        for j in range(self.n_outputs):
            dest[j] = NAN

        for k in range(self.n_node_cols):
            j = self._node_col_indices[k]
            dest[j] = self.sum_total[k] / self.weighted_n_node_samples

    cdef void total_node_value(self, double* dest) noexcept nogil:
        """Compute a single node value for all targets, disregarding y's shape.

        This method is used instead of node_value() in cases where the
        different columns of y are *not* considered as different outputs, being
        usually equivalent to node_value if y were to be flattened, i.e.

            total_node_value(y) == node_value(y.reshape[-1, 1])

        Parameters
        ----------
        dest : double pointer
            The memory address which we will save the node value into.
        """
        cdef SIZE_t k
        dest[0] = 0.0
        for k in range(self.n_node_cols):
            dest[0] += self.sum_total[k]
        dest[0] /= (self.weighted_n_node_samples * self.weighted_n_node_cols)


cdef class AxisSquaredError(AxisRegressionCriterion):
    """Mean squared error impurity criterion.
        MSE = var_left + var_right
    """

    cdef double proxy_impurity_improvement(self) noexcept nogil:
        """Compute a proxy of the impurity reduction.

        This method is used to speed up the search for the best split.
        It is a proxy quantity such that the split that maximizes this value
        also maximizes the impurity improvement. It neglects all constant terms
        of the impurity decrease for a given split.

        The absolute impurity improvement is only computed by the
        impurity_improvement method once the best split has been found.
        The MSE proxy is derived from
            sum_{i left}(y_i - y_pred_L)^2 + sum_{i right}(y_i - y_pred_R)^2
            = sum(y_i^2) - n_L * mean_{i left}(y_i)^2 - n_R * mean_{i right}(y_i)^2
        Neglecting constant terms, this gives:
            - 1/n_L * sum_{i left}(y_i)^2 - 1/n_R * sum_{i right}(y_i)^2

        The improvement on the columns axis is always 0, so we disregard it.
        """
        cdef:
            SIZE_t k
            double proxy_impurity_left = 0.0
            double proxy_impurity_right = 0.0
            double wj = 1.0

        for k in range(self.n_node_cols):
            if self.col_weights is not None:
                wj = self.col_weights[self._node_col_indices[k]]

            proxy_impurity_left += wj * self.sum_left[k] * self.sum_left[k]
            proxy_impurity_right += wj * self.sum_right[k] * self.sum_right[k]

        return (proxy_impurity_left / self.weighted_n_left +
                proxy_impurity_right / self.weighted_n_right)

    cdef void node_axes_impurities(
        self,
        double* rows_impurity,
        double* cols_impurity,
    ) noexcept nogil:
        """Evaluate the impurity of the current node in both directions.
        Evaluate the MSE criterion as impurity of the current node,
        i.e. the impurity of sample_indices[start:end]. The smaller the impurity the
        better.
        """
        # TODO: cache output as we do with self.axes_children_impurities()?
        cdef:
            SIZE_t k
            double wj = 1.0

        rows_impurity[0] = self.sq_sum_total / self.weighted_n_node_samples
        cols_impurity[0] = self.sq_sum_total / self.weighted_n_node_cols

        for k in range(self.n_node_cols):
            if self.col_weights is not None:
                wj = self.col_weights[self._node_col_indices[k]]

            rows_impurity[0] -= (
                wj * (self.sum_total[k] / self.weighted_n_node_samples) ** 2.0
            )

        cols_impurity[0] -= (
            self.sum_sq_row_sums / self.weighted_n_node_cols ** 2.0
        )

        rows_impurity[0] /= self.weighted_n_node_cols
        cols_impurity[0] /= self.weighted_n_node_samples

    cdef void children_axes_impurities(
        self,
        double* rows_impurity_left,
        double* rows_impurity_right,
        double* cols_impurity_left,
        double* cols_impurity_right,
    ) noexcept nogil:
        """Evaluate the impurity in children nodes, in both directions.
        i.e. the impurity of the left child (sample_indices[start:pos]) and the
        impurity the right child (sample_indices[pos:end]).
        """
        if self.pos == self._cached_pos:
            rows_impurity_left[0] = self._cached_rows_impurity_left
            rows_impurity_right[0] = self._cached_rows_impurity_right
            cols_impurity_left[0] = self._cached_cols_impurity_left
            cols_impurity_right[0] = self._cached_cols_impurity_right
            return

        cdef:
            const DOUBLE_t[:] sample_weight = self.sample_weight
            const SIZE_t[:] sample_indices = self.sample_indices
            SIZE_t pos = self.pos
            SIZE_t start = self.start

            const DOUBLE_t[:] col_weights = self.col_weights
            SIZE_t[::1] _node_col_indices = self._node_col_indices
            SIZE_t start_col = self.start_col
            SIZE_t end_col = self.end_col
            SIZE_t n_node_cols = self.n_node_cols

            DOUBLE_t y_ij

            double sq_sum_left = 0.0
            double sq_sum_right
            double sum_sq_row_sums_left = 0.0
            double row_sum
            double sum_sq_col_sums_left = 0.0
            double sum_sq_col_sums_right = 0.0

            SIZE_t i, j, p, k
            DOUBLE_t wi = 1.0, wj = 1.0

        for p in range(start, pos):
            i = sample_indices[p]
            if sample_weight is not None:
                wi = sample_weight[i]

            row_sum = 0.0

            for k in range(n_node_cols):
                j = _node_col_indices[k]
                if col_weights is not None:
                    wj = col_weights[j]

                y_ij = self.y_[i, j]
                sq_sum_left += wi * wj * y_ij * y_ij
                row_sum += wj * y_ij
            
            sum_sq_row_sums_left += wi * row_sum * row_sum

        sq_sum_right = self.sq_sum_total - sq_sum_left
        sum_sq_row_sums_right = self.sum_sq_row_sums - sum_sq_row_sums_left

        rows_impurity_left[0] = sq_sum_left / self.weighted_n_left
        rows_impurity_right[0] = sq_sum_right / self.weighted_n_right
        cols_impurity_left[0] = sq_sum_left / self.weighted_n_node_cols
        cols_impurity_right[0] = sq_sum_right / self.weighted_n_node_cols

        for k in range(self.n_node_cols):
            if col_weights is not None:
                wj = col_weights[_node_col_indices[k]]
            sum_sq_col_sums_left += wj * self.sum_left[k] * self.sum_left[k]
            sum_sq_col_sums_right += wj * self.sum_right[k] * self.sum_right[k]
        
        rows_impurity_left[0] -= (
            sum_sq_col_sums_left / self.weighted_n_left ** 2.0
        )
        rows_impurity_right[0] -= (
            sum_sq_col_sums_right / self.weighted_n_right ** 2.0
        )
        cols_impurity_left[0] -= (
            sum_sq_row_sums_left / self.weighted_n_node_cols ** 2.0
        )
        cols_impurity_right[0] -= (
            sum_sq_row_sums_right / self.weighted_n_node_cols ** 2.0
        )

        rows_impurity_left[0] /= self.weighted_n_node_cols
        rows_impurity_right[0] /= self.weighted_n_node_cols
        cols_impurity_left[0] /= self.weighted_n_left
        cols_impurity_right[0] /= self.weighted_n_right

        self._cached_pos = self.pos
        self._cached_rows_impurity_left = rows_impurity_left[0]
        self._cached_rows_impurity_right = rows_impurity_right[0]
        self._cached_cols_impurity_left = cols_impurity_left[0]
        self._cached_cols_impurity_right = cols_impurity_right[0]

    cdef double _proxy_improvement_factor(self) noexcept nogil:
        """If improvement = proxy_improvement / a + b, this method returns a.

        This is useful when defining proxy impurity improvements for
        compositions of Criterion objects.
        """
        return 2.0 * self.weighted_n_cols * self.weighted_n_samples


cdef class AxisCriterionGSO(AxisCriterion):
    # TODO: when summing, should we divide each term to avoid overflow?

    def __cinit__(self, SIZE_t n_outputs, SIZE_t n_samples, *args, **kwargs):
        self._y_row_sums = np.empty(n_samples, dtype='float64', order='C')
        self._node_col_indices = np.empty(n_outputs, dtype=np.intp, order='C')

    cdef int axis_init(
        self,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight,
        const DOUBLE_t[:] col_weights,
        const SIZE_t[:] sample_indices,
        const SIZE_t[:] col_indices,
        double weighted_n_samples,
        double weighted_n_cols,
        SIZE_t start,
        SIZE_t end,
        SIZE_t start_col,
        SIZE_t end_col,
    ) except -1 nogil:

        self._cached_pos = SIZE_MAX  # Important to reset cached values.
        self.init_columns(
            col_weights=col_weights,
            weighted_n_cols=weighted_n_cols,
            col_indices=col_indices,
            start_col=start_col,
            end_col=end_col,
        )
        # Initialize fields
        self.y_ = y
        self.sample_weight = sample_weight
        self.sample_indices = sample_indices
        self.start = start
        self.end = end
        self.n_node_samples = end - start
        self.weighted_n_samples = weighted_n_samples

        cdef:
            SIZE_t[::1] _node_col_indices = self._node_col_indices
            SIZE_t n_node_cols = self.n_node_cols
            SIZE_t i, j, p, k
            DOUBLE_t y_ij
            DOUBLE_t wi = 1.0, wj = 1.0
            DOUBLE_t row_sum

        self.sq_sum_total = 0.0
        self.weighted_n_node_samples = 0.0
        self.sum_total = 0.0

        for p in range(start, end):
            i = sample_indices[p]
            if sample_weight is not None:
                wi = sample_weight[i]

            self.weighted_n_node_samples += wi
            row_sum = 0.0

            for k in range(n_node_cols):
                j = _node_col_indices[k]
                if col_weights is not None:
                    wj = col_weights[j]

                y_ij = y[i, j]

                row_sum += wj * y_ij
                self.sq_sum_total += wi * wj * y_ij * y_ij

            row_sum *= wi
            self._y_row_sums[i] = row_sum
            self.sum_total += row_sum

        # Reset to pos=start
        self.reset()
        return 0

    cdef int reset(self) except -1 nogil:
        """Reset the criterion at pos=start."""
        self.sum_left = 0.0
        self.sum_right = self.sum_total

        self.weighted_n_left = 0.0
        self.weighted_n_right = self.weighted_n_node_samples
        self.pos = self.start
        return 0

    cdef int reverse_reset(self) except -1 nogil:
        """Reset the criterion at pos=end."""
        self.sum_left = self.sum_total
        self.sum_right = 0.0

        self.weighted_n_left = self.weighted_n_node_samples
        self.weighted_n_right = 0.0
        self.pos = self.end
        return 0

    cdef int update(self, SIZE_t new_pos) except -1 nogil:
        """Updated statistics by moving sample_indices[pos:new_pos] to the left."""
        cdef const DOUBLE_t[:] sample_weight = self.sample_weight
        cdef const SIZE_t[:] sample_indices = self.sample_indices

        cdef SIZE_t pos = self.pos
        cdef SIZE_t end = self.end
        cdef SIZE_t i
        cdef SIZE_t p
        cdef DOUBLE_t w = 1.0

        # Update statistics up to new_pos
        #
        # Given that
        #           sum_left[x] +  sum_right[x] = sum_total[x]
        # and that sum_total is known, we are going to update
        # sum_left from the direction that require the least amount
        # of computations, i.e. from pos to new_pos or from end to new_pos.
        if (new_pos - pos) <= (end - new_pos):
            for p in range(pos, new_pos):
                i = sample_indices[p]

                if sample_weight is not None:
                    w = sample_weight[i]

                self.sum_left += self._y_row_sums[i]  # w is included
                self.weighted_n_left += w
        else:
            self.reverse_reset()

            for p in range(end - 1, new_pos - 1, -1):
                i = sample_indices[p]

                if sample_weight is not None:
                    w = sample_weight[i]

                self.sum_left -= self._y_row_sums[i]  # w is included
                self.weighted_n_left -= w

        self.weighted_n_right = (
            self.weighted_n_node_samples - self.weighted_n_left
        )
        self.sum_right = self.sum_total - self.sum_left

        self.pos = new_pos
        return 0

    cdef void node_value(self, double* dest) noexcept nogil:
        dest[0] = self.sum_total / (
            self.weighted_n_node_samples
            * self.weighted_n_node_cols
        )

    cdef void total_node_value(self, double* dest) noexcept nogil:
        """Compute a single node value for all targets, disregarding y's shape.

        This method is used instead of node_value() in cases where the
        different columns of y are *not* considered as different outputs, being
        usually equivalent to node_value if y were to be flattened, i.e.

            total_node_value(y) == node_value(y.reshape[-1, 1])

        Parameters
        ----------
        dest : double pointer
            The memory address which we will save the node value into.
        """
        self.node_value(dest)
        
    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double impurity_left,
        double impurity_right,
    ) noexcept nogil:
        return (
            (self.weighted_n_node_samples / self.weighted_n_samples)
            * (self.weighted_n_node_cols / self.weighted_n_cols)
            * (
                impurity_parent
                - impurity_right
                    * (self.weighted_n_right / self.weighted_n_node_samples)
                - impurity_left
                    * (self.weighted_n_left / self.weighted_n_node_samples)
            )
        )
    
    cdef double node_impurity(self) noexcept nogil:
        pass

    cdef void children_impurity(
        self,
        double* impurity_left,
        double* impurity_right,
    ) noexcept nogil:
        pass


cdef class AxisSquaredErrorGSO(AxisCriterionGSO):
    """Mean squared error impurity criterion.
        MSE = var_left + var_right
    """
    cdef double node_impurity(self) noexcept nogil:
        cdef double total_weighted_n_node_samples = (
            self.weighted_n_node_samples * self.weighted_n_node_cols
        )
        return (
            self.sq_sum_total / total_weighted_n_node_samples
            - (self.sum_total / total_weighted_n_node_samples) ** 2
        )

    cdef double proxy_impurity_improvement(self) noexcept nogil:
        """Compute a proxy of the impurity reduction.

        This method is used to speed up the search for the best split.
        It is a proxy quantity such that the split that maximizes this value
        also maximizes the impurity improvement. It neglects all constant terms
        of the impurity decrease for a given split.

        The absolute impurity improvement is only computed by the
        impurity_improvement method once the best split has been found.
        """
        return (
            self.sum_left * self.sum_left / self.weighted_n_left
            + self.sum_right * self.sum_right / self.weighted_n_right
        )
    
    cdef double _proxy_improvement_factor(self) noexcept nogil:
        """If improvement = proxy_improvement / a + b, this method returns a.

        This is useful when defining proxy impurity improvements for
        compositions of Criterion objects.
        """
        return (
            self.weighted_n_cols
            * self.weighted_n_node_cols
            * self.weighted_n_samples
        )

    cdef void children_impurity(
        self,
        double* impurity_left,
        double* impurity_right,
    ) noexcept nogil: 
        cdef:
            const DOUBLE_t[:] sample_weight = self.sample_weight
            const SIZE_t[:] sample_indices = self.sample_indices
            SIZE_t pos = self.pos
            SIZE_t start = self.start

            const DOUBLE_t[:] col_weights = self.col_weights
            SIZE_t[::1] _node_col_indices = self._node_col_indices
            SIZE_t n_node_cols = self.n_node_cols

            double sq_sum_left = 0.0
            double sq_sum_right
            double total_weighted_n_left
            double total_weighted_n_right

            SIZE_t i, j, p, k
            DOUBLE_t y_ij
            DOUBLE_t wi = 1.0, wj = 1.0

        # Obtain sq_sum_left 
        for p in range(start, pos):
            i = sample_indices[p]
            if sample_weight is not None:
                wi = sample_weight[i]

            for k in range(n_node_cols):
                j = _node_col_indices[k]
                if col_weights is not None:
                    wj = col_weights[j]

                y_ij = self.y_[i, j]
                sq_sum_left += wi * wj * y_ij * y_ij
            
        sq_sum_right = self.sq_sum_total - sq_sum_left

        total_weighted_n_left = (
            self.weighted_n_left * self.weighted_n_node_cols
        )
        total_weighted_n_right = (
            self.weighted_n_right * self.weighted_n_node_cols
        )

        impurity_left[0] = (
            sq_sum_left / total_weighted_n_left
            - (self.sum_left / total_weighted_n_left) ** 2
        )
        impurity_right[0] = (
            sq_sum_right / total_weighted_n_right
            - (self.sum_right / total_weighted_n_right) ** 2
        )


cdef class AxisFriedmanGSO(AxisSquaredErrorGSO):
    cdef double proxy_impurity_improvement(self) noexcept nogil:
        """Compute a proxy of the impurity reduction.

        This method is used to speed up the search for the best split.
        It is a proxy quantity such that the split that maximizes this value
        also maximizes the impurity improvement. It neglects all constant terms
        of the impurity decrease for a given split.

        The absolute impurity improvement is only computed by the
        impurity_improvement method once the best split has been found.
        """
        cdef double diff = 0.0

        diff = (
            self.weighted_n_right * self.sum_left
            - self.weighted_n_left * self.sum_right
        )

        return diff * diff / (self.weighted_n_left * self.weighted_n_right)

    cdef double impurity_improvement(
        self,
        double impurity_parent,
        double impurity_left,
        double impurity_right
    ) noexcept nogil:
        # Note: none of the arguments are used here
        cdef double diff

        diff = (
            self.weighted_n_right * self.sum_left
            - self.weighted_n_left * self.sum_right
        )

        return (
            diff * diff
            / (
                self.weighted_n_left
                * self.weighted_n_right
                * self.weighted_n_node_samples
                * self.weighted_n_node_cols  # In GMO this factor is squared
            )
        )

    cdef double _proxy_improvement_factor(self) noexcept nogil:
        """If improvement = proxy_improvement / a + b, this method returns a.

        This is useful when defining proxy impurity improvements for
        compositions of Criterion objects.
        """
        return self.weighted_n_node_cols * self.weighted_n_node_samples


cdef class AxisClassificationCriterion(AxisCriterion):
    """Abstract criterion for classification."""

    def __cinit__(
        self,
        SIZE_t n_outputs,
        cnp.ndarray[SIZE_t, ndim=1] n_classes,
    ):
        """Initialize attributes for this criterion.
        Parameters
        ----------
        n_outputs : SIZE_t
            The number of targets, the dimensionality of the prediction
        n_classes : numpy.ndarray, dtype=SIZE_t
            The number of unique classes in each target
        """
        self.start = 0
        self.pos = 0
        self.end = 0

        self.n_outputs = n_outputs
        self.n_samples = 0
        self.n_node_samples = 0
        self.weighted_n_node_samples = 0.0
        self.weighted_n_left = 0.0
        self.weighted_n_right = 0.0

        self.n_classes = np.empty(n_outputs, dtype=np.intp)

        cdef SIZE_t k = 0
        cdef SIZE_t max_n_classes = 0

        # For each target, set the number of unique classes in that target,
        # and also compute the maximal stride of all targets
        for k in range(n_outputs):
            self.n_classes[k] = n_classes[k]

            if n_classes[k] > max_n_classes:
                max_n_classes = n_classes[k]

        self.max_n_classes = max_n_classes

        # Count labels for each output
        self.sum_total = np.zeros((n_outputs, max_n_classes), dtype=np.float64)
        self.sum_left = np.zeros((n_outputs, max_n_classes), dtype=np.float64)
        self.sum_right = np.zeros((n_outputs, max_n_classes), dtype=np.float64)

        self._node_col_indices = np.empty(n_outputs, dtype=np.intp, order='C')

    def __reduce__(self):
        return (type(self),
                (self.n_outputs, np.asarray(self.n_classes)), self.__getstate__())

    cdef int axis_init(
        self,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight,
        const DOUBLE_t[:] col_weights,
        const SIZE_t[:] sample_indices,
        const SIZE_t[:] col_indices,
        double weighted_n_samples,
        double weighted_n_cols,
        SIZE_t start,
        SIZE_t end,
        SIZE_t start_col,
        SIZE_t end_col,
    ) except -1 nogil:
        """Initialize the criterion.
        This initializes the criterion at node sample_indices[start:end] and children
        sample_indices[start:start] and sample_indices[start:end].
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.
        Parameters
        ----------
        y : ndarray, dtype=DOUBLE_t
            The target stored as a buffer for memory efficiency.
        sample_weight : ndarray, dtype=DOUBLE_t
            The weight of each sample stored as a Cython memoryview.
        weighted_n_samples : double
            The total weight of all sample_indices
        sample_indices : ndarray, dtype=SIZE_t
            A mask on the sample_indices. Indices of the sample_indices in X and y we want to use,
            where sample_indices[start:end] correspond to the sample_indices in this node.
        start : SIZE_t
            The first sample to use in the mask
        end : SIZE_t
            The last sample to use in the mask
        """
        self._cached_pos = SIZE_MAX  # Important to reset cached values.
        self.init_columns(
            col_weights=col_weights,
            weighted_n_cols=weighted_n_cols,
            col_indices=col_indices,
            start_col=start_col,
            end_col=end_col,
        )
        self.y_ = y
        self.sample_weight = sample_weight
        self.sample_indices = sample_indices
        self.start = start
        self.end = end
        self.n_node_samples = end - start
        self.weighted_n_samples = weighted_n_samples
        self.weighted_n_node_samples = 0.0

        cdef:
            SIZE_t[::1] _node_col_indices = self._node_col_indices
            SIZE_t n_node_cols = self.n_node_cols
            SIZE_t i, p, k, c, n_classes_k
            DOUBLE_t wi = 1.0

        for k in range(n_node_cols):
            # TODO: in init_columns, store n_classes contiguously like we do
            # with _node_col_indices.
            n_classes_k = self.n_classes[_node_col_indices[k]]
            memset(&self.sum_total[k, 0], 0, n_classes_k * sizeof(double))

        for p in range(start, end):
            i = sample_indices[p]
            # w is originally set to be 1.0, meaning that if no sample weights
            # are given, the default weight of each sample is 1.0.
            if sample_weight is not None:
                wi = sample_weight[i]

            self.weighted_n_node_samples += wi

            for k in range(n_node_cols):
                # Count weighted class frequency for each target
                c = <SIZE_t> self.y_[i, _node_col_indices[k]]
                self.sum_total[k, c] += wi

        # Reset to pos=start
        self.reset()
        return 0

    cdef int reset(self) except -1 nogil:
        """Reset the criterion at pos=start.
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.
        """
        cdef SIZE_t k, n_classes_k
        self.pos = self.start
        self.weighted_n_left = 0.0
        self.weighted_n_right = self.weighted_n_node_samples

        for k in range(self.n_node_cols):
            n_classes_k = self.n_classes[self._node_col_indices[k]]
            memset(&self.sum_left[k, 0], 0, n_classes_k * sizeof(double))
            memcpy(&self.sum_right[k, 0], &self.sum_total[k, 0], n_classes_k * sizeof(double))
        return 0

    cdef int reverse_reset(self) except -1 nogil:
        """Reset the criterion at pos=end.
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.
        """
        cdef SIZE_t k, n_classes_k
        self.pos = self.end
        self.weighted_n_left = self.weighted_n_node_samples
        self.weighted_n_right = 0.0

        for k in range(self.n_node_cols):
            n_classes_k = self.n_classes[self._node_col_indices[k]]
            memset(&self.sum_right[k, 0], 0, n_classes_k * sizeof(double))
            memcpy(&self.sum_left[k, 0],  &self.sum_total[k, 0], n_classes_k * sizeof(double))
        return 0

    cdef int update(self, SIZE_t new_pos) except -1 nogil:
        """Updated statistics by moving sample_indices[pos:new_pos] to the left child.
        Returns -1 in case of failure to allocate memory (and raise MemoryError)
        or 0 otherwise.
        Parameters
        ----------
        new_pos : SIZE_t
            The new ending position for which to move sample_indices from the right
            child to the left child.
        """
        cdef:
            SIZE_t pos = self.pos
            SIZE_t end = self.end
            const SIZE_t[:] sample_indices = self.sample_indices
            const DOUBLE_t[:] sample_weight = self.sample_weight
            const SIZE_t[:] _node_col_indices = self._node_col_indices
            SIZE_t i, p, k, c
            DOUBLE_t wi = 1.0

        # Update statistics up to new_pos
        #
        # Given that
        #   sum_left[x] +  sum_right[x] = sum_total[x]
        # and that sum_total is known, we are going to update
        # sum_left from the direction that require the least amount
        # of computations, i.e. from pos to new_pos or from end to new_po.
        if (new_pos - pos) <= (end - new_pos):
            for p in range(pos, new_pos):
                i = sample_indices[p]
                if sample_weight is not None:
                    wi = sample_weight[i]

                self.weighted_n_left += wi

                for k in range(self.n_node_cols):
                    self.sum_left[k, <SIZE_t>self.y_[i, _node_col_indices[k]]] += wi

        else:
            self.reverse_reset()

            for p in range(end - 1, new_pos - 1, -1):
                i = sample_indices[p]
                if sample_weight is not None:
                    wi = sample_weight[i]

                self.weighted_n_left -= wi

                for k in range(self.n_node_cols):
                    self.sum_left[k, <SIZE_t>self.y_[i, _node_col_indices[k]]] -= wi

        # Update right part statistics
        self.weighted_n_right = self.weighted_n_node_samples - self.weighted_n_left
        for k in range(self.n_node_cols):
            for c in range(self.n_classes[_node_col_indices[k]]):
                self.sum_right[k, c] = self.sum_total[k, c] - self.sum_left[k, c]

        self.pos = new_pos
        return 0

    cdef void node_value(self, double* dest) noexcept nogil:
        """Compute the node value of sample_indices[start:end] and save it into dest.
        Parameters
        ----------
        dest : double pointer
            The memory address which we will save the node value into.
        """
        # FIXME: use missing classes?
        cdef SIZE_t j, k, n_classes_k#, n_missing_classes
        # cdef double[:, :self.max_n_classes:1] dest_  # TRY

        for j in range(self.n_outputs * self.max_n_classes):
            dest[j] = NAN

        for k in range(self.n_node_cols):
            j = self._node_col_indices[k]
            n_classes_k = self.n_classes[j]
            memcpy(
                &dest[j * self.max_n_classes],
                &self.sum_total[k, 0],
                n_classes_k * sizeof(double),
            )
            # n_missing_classes = self.max_n_classes - n_classes
            # memset(&dest[j, n_classes], 0, n_missing_classes * sizeof(double))

    cdef void total_node_value(self, double* dest) noexcept nogil:
        """Compute a single node value for all targets, disregarding y's shape.

        This method is used instead of node_value() in cases where the
        different columns of y are *not* considered as different outputs, being
        usually equivalent to node_value if y were to be flattened, i.e.

            total_node_value(y) == node_value(y.reshape[-1, 1])

        Parameters
        ----------
        dest : double pointer
            The memory address which we will save the node value into.
        """
        cdef SIZE_t j, k, n_classes_k, class_index
        memset(dest, 0, self.max_n_classes * sizeof(double))

        for k in range(self.n_node_cols):
            j = self._node_col_indices[k]
            n_classes_k = self.n_classes[j]

            for class_index in range(n_classes_k):
                dest[class_index] += self.sum_total[k, class_index]


cdef class AxisEntropy(AxisClassificationCriterion):
    r"""Cross Entropy impurity criterion.
    This handles cases where the target is a classification taking values
    0, 1, ... K-2, K-1. If node m represents a region Rm with Nm observations,
    then let
        count_k = 1 / Nm \sum_{x_i in Rm} I(yi = k)
    be the proportion of class k observations in node m.
    The cross-entropy is then defined as
        cross-entropy = -\sum_{k=0}^{K-1} count_k log(count_k)
    """
    cdef void node_axes_impurities(
        self,
        double* rows_impurity,
        double* cols_impurity,
    ) noexcept nogil:
        """Evaluate the impurity of the current node.
        Evaluate the cross-entropy criterion as impurity of the current node,
        i.e. the impurity of sample_indices[start:end]. The smaller the impurity the
        better.
        """
        cdef:
            double transposed_impurity_left
            double transposed_impurity_right
            double entropy = 0.0
            double count_k
            double wj = 1.0
            SIZE_t k, c, j

        for k in range(self.n_node_cols):
            j = self._node_col_indices[k]

            if self.col_weights is not None:
                wj = self.col_weights[j]

            for c in range(self.n_classes[j]):
                count_k = self.sum_total[k, c]
                if count_k > 0.0:
                    count_k /= self.weighted_n_node_samples
                    entropy -= count_k * log(count_k) * wj

        rows_impurity[0] = entropy / self.weighted_n_node_cols

        self._transposed_children_impurity(
            &transposed_impurity_left,
            &transposed_impurity_right,
        )
        cols_impurity[0] = (
            self.weighted_n_left * transposed_impurity_left
            + self.weighted_n_right * transposed_impurity_right
        ) / self.weighted_n_node_samples

    cdef void _rows_children_impurity(
        self,
        double* rows_impurity_left,
        double* rows_impurity_right,
    ) noexcept nogil:
        cdef:
            double entropy_left = 0.0
            double entropy_right = 0.0
            double wj = 1.0
            double count_k
            SIZE_t k, c, j

        for k in range(self.n_node_cols):
            j = self._node_col_indices[k]

            if self.col_weights is not None:
                wj = self.col_weights[j]

            for c in range(self.n_classes[j]):
                count_k = self.sum_left[k, c]
                if count_k > 0.0:
                    count_k /= self.weighted_n_left
                    entropy_left -= wj * count_k * log(count_k)

                count_k = self.sum_right[k, c]
                if count_k > 0.0:
                    count_k /= self.weighted_n_right
                    entropy_right -= wj * count_k * log(count_k)

        rows_impurity_left[0] = entropy_left / self.weighted_n_node_cols
        rows_impurity_right[0] = entropy_right / self.weighted_n_node_cols

    cdef void children_axes_impurities(
        self,
        double* rows_impurity_left,
        double* rows_impurity_right,
        double* cols_impurity_left,
        double* cols_impurity_right,
    ) noexcept nogil:
        """Evaluate the impurity in children nodes.
        i.e. the impurity of the left child (sample_indices[start:pos]) and the
        impurity the right child (sample_indices[pos:end]).
        Parameters
        ----------
        impurity_left : double pointer
            The memory address to save the impurity of the left node
        impurity_right : double pointer
            The memory address to save the impurity of the right node
        """
        self._rows_children_impurity(
            rows_impurity_left,
            rows_impurity_right,
        )
        self._transposed_children_impurity(
            cols_impurity_left,
            cols_impurity_right,
        )

    cdef double proxy_impurity_improvement(self) noexcept nogil:
        cdef double impurity_left, impurity_right

        self._rows_children_impurity(&impurity_left, &impurity_right)

        return (
            - self.weighted_n_left * impurity_left
            - self.weighted_n_right * impurity_right
        )

    cdef void _transposed_children_impurity(
        self,
        double* impurity_left,
        double* impurity_right,
    ) noexcept nogil:
        """Calculates the impurity of the transposed children partitions.
        """
        cdef:
            double* row_counts
            double row_counts_c
            double* current_entropy
            SIZE_t counts_size
            SIZE_t i, j, p, k, c
            double wi = 1.0, wj = 1.0

        row_counts = <double*>malloc(self.max_n_classes * sizeof(double))
        counts_size = self.max_n_classes * sizeof(double)

        impurity_left[0] = 0.0
        impurity_right[0] = 0.0
        current_entropy = impurity_left

        for p in range(self.start, self.end):
            # if current_entropy == impurity_left and p >= self.pos:
            #     current_entropy = impurity_right
            if current_entropy == impurity_left:
                if p >= self.pos:
                    if self.weighted_n_right <= 0.0:
                        break  # Avoid division by zero
                    current_entropy = impurity_right
                elif self.weighted_n_left <= 0.0:
                    continue  # Avoid division by zero

            memset(&row_counts[0], 0, counts_size)

            i = self.sample_indices[p]

            if self.sample_weight is not None:
                wi = self.sample_weight[i]

            for k in range(self.n_node_cols):
                j = self._node_col_indices[k]

                if self.col_weights is not None:
                    wj = self.col_weights[j]
                
                row_counts[<SIZE_t>self.y_[i, j]] += wj
            
            for c in range(self.max_n_classes):
                row_counts_c = row_counts[c]
                if row_counts_c > 0.0:
                    row_counts_c /= self.weighted_n_node_cols
                    current_entropy[0] -= wi * row_counts_c * log(row_counts_c)
        
        if self.weighted_n_left > 0.0:
            impurity_left[0] /= self.weighted_n_left
        if self.weighted_n_right > 0.0:
            impurity_right[0] /= self.weighted_n_right
        
        free(row_counts)


cdef class AxisGini(AxisClassificationCriterion):
    r"""Gini Index impurity criterion.
    This handles cases where the target is a classification taking values
    0, 1, ... K-2, K-1. If node m represents a region Rm with Nm observations,
    then let
        count_k = 1/ Nm \sum_{x_i in Rm} I(yi = k)
    be the proportion of class k observations in node m.
    The Gini Index is then defined as:
        index = \sum_{k=0}^{K-1} count_k (1 - count_k)
              = 1 - \sum_{k=0}^{K-1} count_k ** 2
    """

    cdef void node_axes_impurities(
        self,
        double* rows_impurity, 
        double* cols_impurity, 
    ) noexcept nogil:
        """Evaluate the impurity of the current node.
        Evaluate the Gini criterion as impurity of the current node,
        i.e. the impurity of sample_indices[start:end]. The smaller the impurity the
        better.
        """
        cdef:
            double transposed_impurity_left,
            double transposed_impurity_right,
            double gini = 0.0
            double wj = 1.0
            double sq_count
            double count_k
            SIZE_t k, c, j

        for k in range(self.n_node_cols):
            j = self._node_col_indices[k]
            sq_count = 0.0

            if self.col_weights is not None:
                wj = self.col_weights[j]

            for c in range(self.n_classes[j]):
                count_k = self.sum_total[k, c]
                sq_count += count_k * count_k

            gini += wj * sq_count / (self.weighted_n_node_samples *
                                      self.weighted_n_node_samples)

        rows_impurity[0] = 1.0 - gini / self.weighted_n_node_cols

        self._transposed_children_impurity(
            &transposed_impurity_left,
            &transposed_impurity_right,
        )
        cols_impurity[0] = (
            self.weighted_n_left * transposed_impurity_left
            + self.weighted_n_right * transposed_impurity_right
        ) / self.weighted_n_node_samples

    cdef void _rows_children_impurity(
        self,
        double* rows_impurity_left,
        double* rows_impurity_right,
    ) noexcept nogil:
        cdef:
            double gini_left = 0.0
            double gini_right = 0.0
            double sq_count_left
            double sq_count_right
            double count_k
            double wj = 1.0
            SIZE_t k, c, j

        for k in range(self.n_node_cols):
            j = self._node_col_indices[k]
            sq_count_left = 0.0
            sq_count_right = 0.0

            if self.col_weights is not None:
                wj = self.col_weights[j]

            for c in range(self.n_classes[j]):
                count_k = self.sum_left[k, c]
                sq_count_left += count_k * count_k

                count_k = self.sum_right[k, c]
                sq_count_right += count_k * count_k

            gini_left += wj * sq_count_left / (self.weighted_n_left *
                                                self.weighted_n_left)

            gini_right += wj * sq_count_right / (self.weighted_n_right *
                                                  self.weighted_n_right)

        rows_impurity_left[0] = 1.0 - gini_left / self.weighted_n_node_cols
        rows_impurity_right[0] = 1.0 - gini_right / self.weighted_n_node_cols

    # TODO: cache?
    cdef void children_axes_impurities(
        self,
        double* rows_impurity_left,
        double* rows_impurity_right,
        double* cols_impurity_left,
        double* cols_impurity_right,
    ) noexcept nogil:
        """Evaluate the impurity in children nodes.
        i.e. the impurity of the left child (sample_indices[start:pos]) and the
        impurity the right child (sample_indices[pos:end]).
        Parameters
        ----------
        impurity_left : double pointer
            The memory address to save the impurity of the left node
        impurity_right : double pointer
            The memory address to save the impurity of the right node
        """
        self._rows_children_impurity(
            rows_impurity_left,
            rows_impurity_right,
        )
        self._transposed_children_impurity(
            cols_impurity_left,
            cols_impurity_right,
        )

    cdef double proxy_impurity_improvement(self) noexcept nogil:
        cdef double impurity_left, impurity_right

        self._rows_children_impurity(&impurity_left, &impurity_right)

        return (
            - self.weighted_n_left * impurity_left
            - self.weighted_n_right * impurity_right
        )

    cdef void _transposed_children_impurity(
        self,
        double* impurity_left,
        double* impurity_right,
    ) noexcept nogil:
        """Calculates the impurity of the transposed children partitions.
        """
        cdef:
            double* row_counts
            double row_counts_c
            double* current_sq_count
            SIZE_t counts_size
            SIZE_t i, j, p, k, c
            double wi = 1.0, wj = 1.0

        # TODO: class attribute?
        row_counts = <double*>malloc(self.max_n_classes * sizeof(double))
        counts_size = self.max_n_classes * sizeof(double)

        impurity_left[0] = 0.0
        impurity_right[0] = 0.0
        current_sq_count = impurity_left

        for p in range(self.start, self.end):
            if current_sq_count == impurity_left:
                if p >= self.pos:
                    if self.weighted_n_right <= 0.0:
                        break  # Avoid division by zero
                    current_sq_count = impurity_right
                elif self.weighted_n_left <= 0.0:
                    continue  # Avoid division by zero

            memset(&row_counts[0], 0, counts_size)

            i = self.sample_indices[p]

            if self.sample_weight is not None:
                wi = self.sample_weight[i]

            for k in range(self.n_node_cols):
                j = self._node_col_indices[k]

                if self.col_weights is not None:
                    wj = self.col_weights[j]
                
                row_counts[<SIZE_t>self.y_[i, j]] += wj
            
            for c in range(self.max_n_classes):
                row_counts_c = row_counts[c]
                if row_counts_c > 0.0:
                    row_counts_c /= self.weighted_n_node_cols
                    current_sq_count[0] += wi * row_counts_c * row_counts_c
        
        if self.weighted_n_left > 0.0:
            impurity_left[0] = 1.0 - impurity_left[0] / self.weighted_n_left
        if self.weighted_n_right > 0.0:
            impurity_right[0] = 1.0 - impurity_right[0] / self.weighted_n_right

        free(row_counts)
